/* 
 * Implémentation des fonctions natives
 * pour les API java.io.*
 */

#include <u.h>
#include <libc.h>
#include <lock.h>
#include <auth.h>
#include <fcall.h>

#include "java_const.h"
#include "runtime.h"
#include "fns.h"
#include "native.h"

#include "api_io.h"

/* *************************************************************
 * 
 * JAVA.IO.FILEDESCRIPTOR
 *
 */

Object *
java_io_FileDescriptor_initSystemFD(arg_list args)
{
	Object *	file_desc 	= arg_value(args, Object *);
	int  		desc		= arg_value(args, int);
	
	API_JavaIoFileDescriptor *fd = U(API_JavaIoFileDescriptor, file_desc);

	fd->fd = desc;

	return file_desc;
}

int
java_io_FileDescriptor_valid(arg_list args)
{
	Object *	file_desc 	= arg_value(args, Object *);
	
	API_JavaIoFileDescriptor *fd = U(API_JavaIoFileDescriptor, file_desc);

	return (fd->fd >= 0) ? 1 : 0;
}

/* *************************************************************
 * 
 * JAVA.IO.FILE
 *
 */

int
java_io_File_exists0(arg_list args)
{
	Object *this = arg_value(args, Object *);
		
	char path[1024];
	API_JavaIoFile *file = U(API_JavaIoFile, this);

	if( !file->path ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return 0;
	}

	javaString2CString(file->path, path, sizeof path);
	
	return (access(path, 0) < 0) ? 0 : 1;
}

int
java_io_File_canRead0(arg_list args)
{
	Object *this = arg_value(args, Object *);
		
	char path[1024];
	API_JavaIoFile *file = U(API_JavaIoFile, this);

	if( !file->path ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return 0;
	}

	javaString2CString(file->path, path, sizeof path);
	
	return (access(path, 4) < 0) ? 0 : 1;
}

int
java_io_File_canWrite0(arg_list args)
{
	Object *this = arg_value(args, Object *);
		
	char path[1024];
	API_JavaIoFile *file = U(API_JavaIoFile, this);

	if( !file->path ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return 0;
	}

	javaString2CString(file->path, path, sizeof path);
	
	return (access(path, 2) < 0) ? 0 : 1;
}

int
java_io_File_isFile0(arg_list args)
{
	Object *this = arg_value(args, Object *);
		
	char path[1024];
	char ap[DIRLEN]; Dir edir;
	API_JavaIoFile *file = U(API_JavaIoFile, this);

	if( !file->path ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return 0;
	}

	javaString2CString(file->path, path, sizeof path);
	
	if( stat(path, ap) < 0 )
		return 0;
	memset(&edir, 0, sizeof(Dir));
	convM2D(ap, &edir);
	
	return (edir.mode&CHDIR) ? 0 : 1;
}

int
java_io_File_isDirectory0(arg_list args)
{
	Object *this = arg_value(args, Object *);
		
	char path[1024];
	char ap[DIRLEN]; Dir edir;
	API_JavaIoFile *file = U(API_JavaIoFile, this);

	if( !file->path ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return 0;
	}

	javaString2CString(file->path, path, sizeof path);
	
	if( stat(path, ap) < 0 )
		return 0;
	memset(&edir, 0, sizeof(Dir));
	convM2D(ap, &edir);
	
	return (edir.mode&CHDIR) ? 1 : 0;
}

vlong
java_io_File_length0(arg_list args)
{
	Object *this = arg_value(args, Object *);
		
	char path[1024];
	char ap[DIRLEN]; Dir edir;
	API_JavaIoFile *file = U(API_JavaIoFile, this);

	if( !file->path ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return 0;
	}

	javaString2CString(file->path, path, sizeof path);
	
	if( stat(path, ap) < 0 )
		return 0;
	memset(&edir, 0, sizeof(Dir));
	convM2D(ap, &edir);
	
	return edir.length;
}

int
java_io_File_mkdir0(arg_list args)
{
	Object *this = arg_value(args, Object *);
		
	char path[1024];
	API_JavaIoFile *file = U(API_JavaIoFile, this);
	
	if( !file->path ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return 0;
	}

	javaString2CString(file->path, path, sizeof path);
	
	return (create(path, OREAD, CHDIR|0777L) < 0) ? 0 : 1;
}

int
java_io_File_delete0(arg_list args)
{
	Object *this = arg_value(args, Object *);
		
	char path[1024];
	API_JavaIoFile *file = U(API_JavaIoFile, this);

	if( !file->path ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return 0;
	}

	javaString2CString(file->path, path, sizeof path);
	
	return (remove(path) < 0) ? 0 : 1;
}

int
java_io_File_renameTo0(arg_list args)
{
	Object *this = arg_value(args, Object *);
	Object *dest = arg_value(args, Object *);
		
	char path[1024], dpath[1024];
	char ap[DIRLEN]; Dir edir;
	API_JavaIoFile *file_org = U(API_JavaIoFile, this);
	API_JavaIoFile *file_dst = U(API_JavaIoFile, dest);

	if( !file_org->path || !dest ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return 0;
	}

	javaString2CString(file_org->path, path, sizeof path);	
	if( stat(path, ap) < 0 )
		return 0;
	memset(&edir, 0, sizeof(Dir));
	convM2D(ap, &edir);

	javaString2CString(file_dst->path, dpath, sizeof dpath);
	if( strlen(dpath) > NAMELEN )
		return 0;
	strcpy(edir.name, dpath);
	convD2M(&edir, ap);

	return (wstat(path, ap) < 0) ? 0 : 1;
}

int
java_io_File_isAbsolute(arg_list args)
{
	Object *this = arg_value(args, Object *);
		
	char path[1024];
	API_JavaIoFile *file = U(API_JavaIoFile, this);

	if( !file->path ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return 0;
	}

	javaString2CString(file->path, path, sizeof path);
	
	return (path[0] != '/') ? 0 : 1;
}

Object *
java_io_File_list0(arg_list args)
{
	Object *this = arg_value(args, Object *);
		
	char path[1024];
	int *v;
	int i, n, fd, count;
	Dir buf[25];
	struct list { Object *o; struct list *next; };
	struct list head, *pt, *p;
	Object *s, *tstring;
	API_JavaIoFile *file = U(API_JavaIoFile, this);

	if( !file->path ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return 0;
	}

	javaString2CString(file->path, path, sizeof path);

	/* ouvre le répertoire */

	fd = open(path, OREAD);
	if( fd < 0 )
		return 0;

	/* parcours du répertoire et construit une liste des fichiers */
	
	count = 0;
	head.next = 0;

	while( (n = dirread(fd, buf, sizeof(buf))) > 0 ){
		n /= sizeof(Dir);
		for(i=0 ; i < n ; i++){
			s = makeJavaString(buf[i].name);
			pt = (struct list *)sysMalloc(sizeof(struct list));
			pt->o = s;
			pt->next = head.next;
			head.next = pt;
			count += 1;
		}
	}
	close(fd);

	/* construction du tableau de String */

	tstring = newObjectArray(getClass("java/lang/String"), count);
	
	pt = head.next;
	v = tstring->vars;
	while( pt ){
		*v++ = (int)pt->o;
		p = pt;
		pt = pt->next;
		free(p);
	}

	return tstring;
}

vlong
java_io_File_lastModified0(arg_list args)
{
	Object *this = arg_value(args, Object *);
		
	char path[1024];
	char ap[DIRLEN]; Dir edir;
	API_JavaIoFile *file = U(API_JavaIoFile, this);

	if( !file->path ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return 0;
	}

	javaString2CString(file->path, path, sizeof path);
	
	if( stat(path, ap) < 0 )
		return 0;
	memset(&edir, 0, sizeof(Dir));
	convM2D(ap, &edir);
	
	return edir.mtime;
}


/* *************************************************************
 * 
 * JAVA.IO.RANDOMACCESSFILE
 *
 */


void
java_io_RandomAccessFile_open(arg_list args)
{
	Object *	this	= arg_value(args, Object *);
	Object * 	str		= arg_value(args, Object *);
	int 		write	= arg_value(args, int);
	
	char name[1024];
	int fd;
	
	API_JavaIoRandomAccessFile *raf = U(API_JavaIoRandomAccessFile, this);
	API_JavaIoFileDescriptor *file_desc = U(API_JavaIoFileDescriptor, raf->fd);
	
	/*if( !file_desc || !str ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return;
	}*/

	javaString2CString(str, name, sizeof name);
	
	if( write ){
		if( access(name, 0) < 0 ){
			if( create(name, ORDWR, 0777) < 0 ){
				throwException(getCurrentFrame(), "java/io/IOException");
				return;
			}
		}
		fd = open(name, ORDWR);
	}else
		fd = open(name, OREAD);

	file_desc->fd = fd;
	if( fd < 0 ){
		throwException(getCurrentFrame(), "java/io/IOException");
		return;
	}

	return;
}

int
java_io_RandomAccessFile_read(arg_list args)
{
	Object *	this	= arg_value(args, Object *);
	
	uchar byte;
	int r;
	API_JavaIoRandomAccessFile *raf = U(API_JavaIoRandomAccessFile, this);
	API_JavaIoFileDescriptor *file_desc = U(API_JavaIoFileDescriptor, raf->fd);
	
	/*if( !file_desc ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return -1;
	}*/
	
	if( (r = read(file_desc->fd, &byte, 1)) <= 0 ){
		if( !r ) /* EOF */
			return -1;
		throwException(getCurrentFrame(), "java/io/IOException");
		return -1;
	}

	return byte;
}

int
java_io_RandomAccessFile_readBytes(arg_list args)
{
	Object *	this		= arg_value(args, Object *);
	Object *	tab_bytes 	= arg_value(args, Object *);
	int 		offset		= arg_value(args, int);
	int 		len			= arg_value(args, int);
	
	int r;
	
	API_JavaIoRandomAccessFile *raf = U(API_JavaIoRandomAccessFile, this);
	API_JavaIoFileDescriptor *file_desc = U(API_JavaIoFileDescriptor, raf->fd);
	
	/*if( !file_desc ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return -1;
	}*/
	
	if( (r = read(file_desc->fd, (char *)tab_bytes->vars + offset, len)) < 0 )
		throwException(getCurrentFrame(), "java/io/IOException");

	return (r > 0) ? r : -1;
}

void
java_io_RandomAccessFile_write(arg_list args)
{
	Object *	this	= arg_value(args, Object *);
	int 		byte	= arg_value(args, int);

	char buf[1];
	
	API_JavaIoRandomAccessFile *raf = U(API_JavaIoRandomAccessFile, this);
	API_JavaIoFileDescriptor *file_desc = U(API_JavaIoFileDescriptor, raf->fd);
	
	/*if( !file_desc ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return;
	}*/
	
	buf[0] = byte;
	if( write(file_desc->fd, buf, 1) < 0 ){
		throwException(getCurrentFrame(), "java/io/IOException");
		return;
	}

}

void
java_io_RandomAccessFile_writeBytes(arg_list args)
{
	Object *	this		= arg_value(args, Object *);
	Object *	tab_bytes 	= arg_value(args, Object *);
	int 		offset		= arg_value(args, int);
	int 		len			= arg_value(args, int);
	
	API_JavaIoRandomAccessFile *raf = U(API_JavaIoRandomAccessFile, this);
	API_JavaIoFileDescriptor *file_desc = U(API_JavaIoFileDescriptor, raf->fd);
	
	/*if( !file_desc ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return -1;
	}*/
	
	if( write(file_desc->fd, (char *)tab_bytes->vars + offset, len) < 0 ){
		throwException(getCurrentFrame(), "java/io/IOException");
		return;
	}

}

vlong
java_io_RandomAccessFile_getFilePointer(arg_list args)
{
	Object *	this		= arg_value(args, Object *);
	
	API_JavaIoRandomAccessFile *raf = U(API_JavaIoRandomAccessFile, this);
	API_JavaIoFileDescriptor *file_desc = U(API_JavaIoFileDescriptor, raf->fd);
	
	int cur;

	/*if( !file_desc ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return -1;
	}*/
	
	if( (cur = seek(file_desc->fd, 0L, 1)) < 0 ){
		throwException(getCurrentFrame(), "java/io/IOException");
		return -1;
	}
	
	return (vlong)cur;
}

void
java_io_RandomAccessFile_seek(arg_list args)
{
	Object *	this		= arg_value(args, Object *);
	vlong 		pos			= arg_value(args, vlong);

	API_JavaIoRandomAccessFile *raf = U(API_JavaIoRandomAccessFile, this);
	API_JavaIoFileDescriptor *file_desc = U(API_JavaIoFileDescriptor, raf->fd);
	
	/*if( !file_desc ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return -1;
	}*/
	
	if( seek(file_desc->fd, (long)pos, 0) < 0 ){
		throwException(getCurrentFrame(), "java/io/IOException");
		return;
	}
}

vlong
java_io_RandomAccessFile_length(arg_list args)
{
	Object *	this		= arg_value(args, Object *);
	
	API_JavaIoRandomAccessFile *raf = U(API_JavaIoRandomAccessFile, this);
	API_JavaIoFileDescriptor *file_desc = U(API_JavaIoFileDescriptor, raf->fd);
	
	int cur;

	/*if( !file_desc ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return -1;
	}*/
	
	if( (cur = seek(file_desc->fd, 0L, 2)) < 0 ){
		throwException(getCurrentFrame(), "java/io/IOException");
		return -1;
	}
	
	return (vlong)cur;
}

void
java_io_RandomAccessFile_close(arg_list args)
{
	Object *	this		= arg_value(args, Object *);

	API_JavaIoRandomAccessFile *raf = U(API_JavaIoRandomAccessFile, this);
	API_JavaIoFileDescriptor *file_desc = U(API_JavaIoFileDescriptor, raf->fd);
	
	/*if( !file_desc ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return -1;
	}*/
	
	if( close(file_desc->fd) ){
		throwException(getCurrentFrame(), "java/io/IOException");
		return;
	}
	file_desc->fd = -1;
}

/* *************************************************************
 * 
 * JAVA.IO.FILEINPUTSTREAM
 *
 */


void
java_io_FileInputStream_open(arg_list args)
{
	Object *	this	= arg_value(args, Object *);
	Object * 	str		= arg_value(args, Object *);
	
	char name[1024];
	int fd;
	
	API_JavaIoFileInputStream *raf = U(API_JavaIoFileInputStream, this);
	API_JavaIoFileDescriptor *file_desc = U(API_JavaIoFileDescriptor, raf->fd);
	
	/*if( !file_desc || !str ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return;
	}*/

	javaString2CString(str, name, sizeof name);
	
	fd = open(name, OREAD);
	file_desc->fd = fd;
	if( fd < 0 ){
		throwException(getCurrentFrame(), "java/io/IOException");
		return;
	}

	return;	
}

int
java_io_FileInputStream_read(arg_list args)
{
	Object *	this	= arg_value(args, Object *);
	
	uchar byte;
	int r;
	API_JavaIoFileInputStream *raf = U(API_JavaIoFileInputStream, this);
	API_JavaIoFileDescriptor *file_desc = U(API_JavaIoFileDescriptor, raf->fd);
	
	/*if( !file_desc ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return -1;
	}*/
	
	if( (r = read(file_desc->fd, &byte, 1)) <= 0 ){
		if( !r ) /* EOF */
			return -1;
		throwException(getCurrentFrame(), "java/io/IOException");
		return -1;
	}
	//print("read() byte = %c\n", byte);
	return byte;
}

int
java_io_FileInputStream_readBytes(arg_list args)
{
	Object *	this		= arg_value(args, Object *);
	Object *	tab_bytes 	= arg_value(args, Object *);
	int 		offset		= arg_value(args, int);
	int 		len			= arg_value(args, int);
	
	int r;
	
	API_JavaIoFileInputStream *raf = U(API_JavaIoFileInputStream, this);
	API_JavaIoFileDescriptor *file_desc = U(API_JavaIoFileDescriptor, raf->fd);
	
	/*if( !file_desc ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return -1;
	}*/

	if( (r = read(file_desc->fd, (char *)tab_bytes->vars + offset, len)) < 0 )
		throwException(getCurrentFrame(), "java/io/IOException");

	//print("readBytes() r = %d\n", r);
	return (r > 0) ? r : -1;
}

vlong
java_io_FileInputStream_skip(arg_list args)
{
	Object *	this		= arg_value(args, Object *);
	vlong 		n			= arg_value(args, vlong);

	API_JavaIoFileInputStream *raf = U(API_JavaIoFileInputStream, this);
	API_JavaIoFileDescriptor *file_desc = U(API_JavaIoFileDescriptor, raf->fd);
	vlong cur, ret;

	/*if( !file_desc ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return -1;
	}*/
	
	ret = seek(file_desc->fd, 0L, 1);
	cur = seek(file_desc->fd, (long)n, 1);
	if( cur < 0 ){
		throwException(getCurrentFrame(), "java/io/IOException");
		return 0;
	}
	//print("skip() cur - ret = %lld\n", cur-ret);
	return cur - ret;
}

int
java_io_FileInputStream_available(arg_list args)
{
	Object *	this		= arg_value(args, Object *);

	API_JavaIoFileInputStream *raf = U(API_JavaIoFileInputStream, this);
	API_JavaIoFileDescriptor *file_desc = U(API_JavaIoFileDescriptor, raf->fd);
	long cur, ret, pos;

	/*if( !file_desc ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return -1;
	}*/
	
	cur = seek(file_desc->fd, 0L, 1);
	if( cur < 0 ){
		char buf[1024];
		errstr(buf);
		print("%s\n", buf);
		throwException(getCurrentFrame(), "java/io/IOException");
		return 0;
	}
	ret = seek(file_desc->fd, 0L, 2);
	if( ret < 0 ){
		throwException(getCurrentFrame(), "java/io/IOException");
		return 0;
	}
	pos = seek(file_desc->fd, cur, 0);
	if( pos < 0 ){
		throwException(getCurrentFrame(), "java/io/IOException");
		return 0;
	}
	//print("available() = %d\n", ret - cur);
	return ret - cur;
}

void
java_io_FileInputStream_close(arg_list args)
{
	Object *	this		= arg_value(args, Object *);

	API_JavaIoFileInputStream *raf = U(API_JavaIoFileInputStream, this);
	API_JavaIoFileDescriptor *file_desc = U(API_JavaIoFileDescriptor, raf->fd);
	
	/*if( !file_desc ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return -1;
	}*/
	
	if( close(file_desc->fd) ){
		throwException(getCurrentFrame(), "java/io/IOException");
		return;
	}
	file_desc->fd = -1;
}

/* *************************************************************
 * 
 * JAVA.IO.FILEOUTPUTSTREAM
 *
 */

void
java_io_FileOutputStream_open(arg_list args)
{
	Object *	this	= arg_value(args, Object *);
	Object * 	str		= arg_value(args, Object *);
	
	char name[1024];
	int fd;
	
	API_JavaIoFileOutputStream *raf = U(API_JavaIoFileOutputStream, this);
	API_JavaIoFileDescriptor *file_desc = U(API_JavaIoFileDescriptor, raf->fd);
	
	/*if( !file_desc || !str ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return;
	}*/

	javaString2CString(str, name, sizeof name);

	if( access(name, 0) < 0 ){
		if( create(name, ORDWR, 0777) < 0 ){
			throwException(getCurrentFrame(), "java/io/IOException");
			return;
		}
	}

	fd = open(name, ORDWR);

	file_desc->fd = fd;
	if( fd < 0 ){
		throwException(getCurrentFrame(), "java/io/IOException");
		return;
	}

	return;	
}

void
java_io_FileOutputStream_write(arg_list args)
{
	Object *	this	= arg_value(args, Object *);
	int 		byte	= arg_value(args, int);

	char buf[1];
	
	API_JavaIoFileOutputStream *raf = U(API_JavaIoFileOutputStream, this);
	API_JavaIoFileDescriptor *file_desc = U(API_JavaIoFileDescriptor, raf->fd);
	
	/*if( !file_desc ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return;
	}*/
	
	buf[0] = byte;
	if( write(file_desc->fd, buf, 1) < 0 ){
		throwException(getCurrentFrame(), "java/io/IOException");
		return;
	}

}

void
java_io_FileOutputStream_writeBytes(arg_list args)
{
	Object *	this		= arg_value(args, Object *);
	Object *	tab_bytes 	= arg_value(args, Object *);
	int 		offset		= arg_value(args, int);
	int 		len			= arg_value(args, int);
//	int i;
	API_JavaIoFileOutputStream *raf = U(API_JavaIoFileOutputStream, this);
	API_JavaIoFileDescriptor *file_desc = U(API_JavaIoFileDescriptor, raf->fd);
	
	/*if( !file_desc ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return -1;
	}*/
/*	print("<");
	for(i=0 ; i < len ; i++)
		print("%c ", *((char *)tab_bytes->vars + offset + i));
	print(">\n");
*/
	if( write(file_desc->fd, (char *)tab_bytes->vars + offset, len) < 0 ){
		throwException(getCurrentFrame(), "java/io/IOException");
		return;
	}

}

void
java_io_FileOutputStream_close(arg_list args)
{
	Object *	this		= arg_value(args, Object *);

	API_JavaIoFileOutputStream *raf = U(API_JavaIoFileOutputStream, this);
	API_JavaIoFileDescriptor *file_desc = U(API_JavaIoFileDescriptor, raf->fd);
	
	/*if( !file_desc ){
		throwException(getCurrentFrame(), "java/lang/NullPointerException");
		return -1;
	}*/
	
	if( close(file_desc->fd) ){
		throwException(getCurrentFrame(), "java/io/IOException");
		return;
	}
	file_desc->fd = -1;
}
