/*
 * Classlist
 */

#include <u.h>
#include <libc.h>
#include <lock.h>

#include "java_const.h"
#include "runtime.h"
#include "fns.h"


#define HMAX	31

static char *getClassName(ClassFile *);
static int hashkey(char *);

struct List {
	char *class_name;
	ClassFile *class;
	struct List *next;
};

static struct List *hash[HMAX];

void
addClass(ClassFile *class)
{
	int key;
	struct List *p = (struct List *) sysMalloc(sizeof(struct List));

	if( !p )
		panic("sysMalloc failed addClass()");

	key = hashkey(class->r_name);

	p->class_name = class->r_name;
	p->class = class;
	p->next = hash[key];
	hash[key] = p;
}



ClassFile *
getClass(char *name)
{
	struct List *p = hash[hashkey(name)];
	
	while( p ){
		if( !strcmp(name, p->class_name) )
			return p->class;
		p = p->next;
	}

	return 0;
}

static int
hashkey(char *s)
{
	char *p;
	unsigned int h = 0, g;
     
	for(p=s ; *p ; p++){
		h = (h<<4) + *p;
		if( (g = h & 0xF0000000) ){
			h ^= g>>24;
			h ^= g;
		}
	}
	return h % HMAX;
}

void 
dumpClassArea(void)
{
	struct List *p;
	int i;

	chat("* Class Area *\n");
	for(i=0 ; i < HMAX ; i++){
		if( !hash[i] )
			continue;
		p = hash[i];
		while( p ){
			chat("class %s\n", p->class_name);
			p = p->next;
		}
	}
}

			
static 
char *getClassName(ClassFile *class)
{
	cp_info *c = &class->constant_pool[class->this_class];
	cp_info *cp = &class->constant_pool[c->u.c.name_index];
	char *s = (char *) malloc(cp->u.a.length + 1);

	if( !s )
		panic("getClassName()");

	strncpy(s, (char *)cp->u.a.bytes, cp->u.a.length);
	s[cp->u.a.length] = 0;

	return s;
}