/*
 * classloader.c
 */

ClassFile *	loadClassFileFromZip(char *, zip *, char *);
ClassFile *	loadClassFileFromFs(char *);
void 		makePath(void);

/*
 * classlist.c
 */

void 		addClass(ClassFile *);
ClassFile *	getClass(char *);
void 		dumpClassArea(void);

/* 
 * resolution.c
 */

ClassFile *	loadClass(char *);
void 		linking(ClassFile *);
void 		initialize(ClassFile *);
void 		cp_resolution(ClassFile *, int, Object *);
int 		signatureSize(char *);
int 		argSize(char *);
//void 		GetCpString(char *, int, ClassFile *);
void 		getClassNameFromCP(char *, int, ClassFile *);
void 		getNameFromCP(char *, int, ClassFile *);
void 		getSignatureFromCP(char *, int, ClassFile *);
int 		isSubClassOf(ClassFile *, ClassFile *);
void		getClassFromCP(char *, int, ClassFile *);
int			isClassArray(int, ClassFile *);
int			isImplementOf(ClassFile *, ClassFile *);
int			classArrayType(int, ClassFile *);
ClassFile *	lookupClass(char *);
void 		getSuperClassName(char *, ClassFile *);
int 		isCompatibleObject(ClassFile *, ClassFile *);
int 		isCompatibleArray(Object *, char *);

/* 
 * thread.c
 */

void 		initThread(void);
Object *	getCurrentThread(void);
void		setPriorityThread(Object *, int);
void 		startThread(Object *);
void 		yieldThread(void);
void 		sleepThread(vlong);
void		waitThread(Object *, vlong );
void		notifyThread(Object *object);
void		notifyAllThread(Object *object);
int 		isAliveThread(Object *);
void 		stopThread(Object *, Object *);
void		suspendThread(Object *);
void		resumeThread(Object *);
int 		getLock(Object *);
void 		releaseLock(Object *);
void 		schedule(void);
void		killThread(Object *);

/* 
 * interpreter.c
 */

void 		interpreter(void);
vlong		popLong(Frame *);
void 		executeConstructor(Object *, char *,...);
void		executeVirtual(ClassFile *, Object *, char *, char *,...);
double 		popDouble(Frame *);
void		dumpProfile(void);

/* 
 * object.c
 */

Object *	newObject(ClassFile *);
Object *	newArray(int, int);
Object *	newObjectArray(ClassFile *, int);
Object *	newMultiArray(ClassFile *, int, int, int *);
Object *	newString(ClassFile *);
int 		arrayLength(Object *);
void 		javaString2CString(Object *, char *, int);
Object *	makeJavaString(char *);
void 		setProperty(Object *, char *, char *);
int 		sign2type(char *);
int 		typeSize(int);

/* 
 * native.c 
 */
void 		invokeNative(Frame *, char *, arg_list);

/* 
 * frame.c
 */

StackFrame *initGlobalStack(void);
Frame *		getCurrentFrame(void);
Frame *		newFrameFromName(ClassFile *, char *, char *);
Frame *		createFrameFromName(StackFrame *, ClassFile *, char *, char *);
Frame *		newFrame(MethodBlock *, Frame *);
Frame *		createFrame(StackFrame *, MethodBlock *, Frame *);
void 		freeFrame(Frame *);
Frame *		getCurrentFrame(void);
void 		dumpStackFrame(void);
void 		fillInStackTrace(Object *);
int 		pc2LineCode(MethodBlock *, int);
Frame *		findInvokerFrame(Frame *);

/* 
 * exception.c
 */

void  		throwException(Frame *, char *);

/* 
 * util.c
 */ 

void 		convU2(u2 *);
void 		convU4( u4*);
vlong		atoll(char *);
void *		sysMalloc(int);
void *		sysCalloc(int);
long 		getMiliSec(void);
int 		getNbPath(char *);
void 		getPath(char *, char *, int);
int 		truncName(char *);
void		sysExit(int);
void		point2slash(char *);
void		slash2point(char *);
void		lushr(vlong *, int);
void 		dumpFrame(Frame *);
void 		dumpObject(Object *);
void 		printAccessFlag(int);
void 		dumpString(Object *);

/* 
 * main.c
 */


/* 
 * chat.c
 */

void 		chat(char *, ...);
void 		panic(char *, ...);
void 		mchat(char *, ...);
void 		debug(int indent, char *fmt, ...);

