/*
 * interpreteur
 */

#include <u.h>
#include <libc.h>
#include <lock.h>

#include "java_const.h"
#include "runtime.h"
#include "fns.h"

#include "code_struct.h"
#include "code_enum.h"

static void nativeCall(MethodBlock *, int, int);

#ifdef DBG_1
int prof;
#endif

#ifdef PROFILE
struct {
	int nb;
	ulong time;
} inst_prof[256];
#endif

extern int need_schedule;
extern Frame *current_frame;
extern Object *current_thread;
extern int syn_lock;

#define offset(f,offset)	(f)->pc += (offset)
#define pc2signedshort(pc)  ((((signed char *)(pc))[1] << 8) | (pc)[2])
#define pc2signedlong(pc)  	((((signed char *)(pc))[1] << 24) | ((pc)[2]  << 16) \
			      			| ((pc)[3] << 8) | ((pc)[4]))	

#define TEST_EXCEPTION()	if( f->exception ) break
#define singlePop(f)		(f)->opstack_pt--
#define nPop(f, n)			(f)->opstack_pt -= (n)

void
interpreter(void)
{
	uchar *pc;
	int opcode;
	int index;

	#define f current_frame

#ifdef PROFILE
	long cur_time;
#endif
	for(;;){

		if( need_schedule ){
			//print("%s.%s call schedule()\n", f->class->r_name, f->name);
			schedule();
		}

DBG3(	dumpFrame(f);)

		pc = f->pc;
		opcode = *f->pc;

		if( f->exception ){ 		/* une exception a été lancée  */
			opcode = _athrow;		/* sur la dernière instruction */
DBG1(		chat("receive exception excute athrow.\n"); )
		}else 
		if( opcode == _wide ){
			opcode = *(++f->pc) + 256;
			pc++;
		}

#ifdef PROFILE
		inst_prof[opcode&0xFF].nb += 1; 
		cur_time = getMiliSec();
#endif
DBG2(	debug(prof, "-->%s \n", Instructions[opcode&0xFF].name); )

		switch(opcode){ /* decodage des instructions */

			case _bipush:
				push(f, (char)pc[1]); 
				offset(f, 2);
				break;
			case _sipush:
				push(f, pc2signedshort(pc)); 
				offset(f, 3);
				break;

			case _ldc: {
				cp_info *cp;

				index = pc[1];
				cp = &f->class->constant_pool[index];
				if( cp->tag == CONSTANT_String ){
					if( !f->class->r_resolved[index] ){
						cp_resolution(f->class, index, 0);
						TEST_EXCEPTION();
					}
					push(f, f->class->r_resolved[index]);
				}else
					/* pour les Floats ????????? */
					push(f, cp->u.i.bytes);

				offset(f, 2);
				break;
			}

			case _ldc_w: {
				cp_info *cp;

				index = pc2signedshort(pc);
				cp = &f->class->constant_pool[index];
				if( cp->tag == CONSTANT_String ){
					if( !f->class->r_resolved[index] ){
						cp_resolution(f->class, index, 0);
						TEST_EXCEPTION();
					}
					push(f, f->class->r_resolved[index]);
				}else
					/* pour les Floats ????????? */
					push(f, cp->u.i.bytes);

				offset(f, 3);
				break;
			}
				

			case _ldc2_w: {
				cp_info *cp;
				
				cp = &f->class->constant_pool[pc2signedshort(pc)];
				push(f, cp->u.l.low_bytes);
				push(f, cp->u.l.hight_bytes);
				offset(f, 3);
				break;
			}

			case _aconst_null:
				push(f, 0); /* null object */
				offset(f, 1);
				break;
			case _iconst_m1:
			case _iconst_0: case _iconst_1: case _iconst_2:
			case _iconst_3: case _iconst_4: case _iconst_5:
				push(f, opcode-_iconst_0); offset(f, 1);
				break;
			case _lconst_0: case _lconst_1:
				pushLong(f, opcode-_lconst_0);
				offset(f, 1);
				break;
			case _fconst_0: case _fconst_1: case _fconst_2:
				pushFloat(f, (float)(opcode-_fconst_0)); 
				offset(f, 1);
				break;
			case _dconst_0: case _dconst_1: 
				pushDouble(f, (double)(opcode-_dconst_0)); 
				offset(f, 1);
				break;
		
			case _istore: case _astore:
				setlocal(f, pc[1], pop(f));
				offset(f, 2);
				break;
			case _fstore:
				setlocalFloat(f, pc[1], popFloat(f));
				offset(f, 2);
				break;
			case _lstore: 
				setlocalLong(f, pc[1], popLong(f));
				offset(f, 2);
				break;
			case _dstore:
				setlocalDouble(f, pc[1], popDouble(f));
				offset(f, 2);
				break;

			case _istore_0: case _istore_1: 
			case _istore_2: case _istore_3:
				setlocal(f, opcode-_istore_0, pop(f));
				offset(f, 1);
				break;
			case _astore_0: case _astore_1: 
			case _astore_2: case _astore_3:
				setlocal(f, opcode-_astore_0, pop(f));
				offset(f, 1);
				break;
			case _lstore_0: case _lstore_1:
			case _lstore_2: case _lstore_3:
				setlocalLong(f, opcode-_lstore_0, popLong(f));
				offset(f, 1);
				break;
			case _fstore_0: case _fstore_1: 
			case _fstore_2: case _fstore_3:
				setlocalFloat(f, opcode-_fstore_0, popFloat(f));
				offset(f, 1);
				break;
			case _dstore_0: case _dstore_1:
			case _dstore_2: case _dstore_3:
				setlocalDouble(f, opcode-_dstore_0, popDouble(f));
				offset(f, 1);
				break;

			case _iload: case _aload:
				push(f, getlocal(f, pc[1]));
				offset(f, 2);
				break;
			case _fload:
				pushFloat(f, getlocalFloat(f, pc[1]));
				offset(f, 2);
				break;
			case _lload: 
				pushLong(f, getlocalLong(f, pc[1]));
				offset(f, 2);
				break;
			case _dload:
				pushDouble(f, getlocalDouble(f, pc[1]));
				offset(f, 2);
				break;

			case _iload_0: case _iload_1:
			case _iload_2: case _iload_3:
				push(f, getlocal(f, opcode-_iload_0));
				offset(f, 1);
				break;
			case _fload_0: case _fload_1:
			case _fload_2: case _fload_3:
				pushFloat(f, getlocalFloat(f, opcode-_fload_0));
				offset(f, 1);
				break;
			case _aload_0: case _aload_1:
			case _aload_2: case _aload_3:
				push(f, getlocal(f, opcode-_aload_0));
				offset(f, 1);
				break;
			case _lload_0: case _lload_1:
			case _lload_2: case _lload_3:
				pushLong(f, getlocalLong(f, opcode-_lload_0));
				offset(f, 1);
				break;
			case _dload_0: case _dload_1:
			case _dload_2: case _dload_3:
				pushDouble(f, getlocalDouble(f, opcode-_dload_0));
				offset(f, 1);
				break;

			case _iinc:
				setlocal(f, pc[1], getlocal(f, pc[1]) + (signed char)pc[2]);
				offset(f, 3);
				break;

	/* **********************************************
	 *
	 * Wide instructions
	 *
	 */

			
			case _iload + 256: case _aload + 256:
				push(f, getlocal(f, pc2signedshort(pc)));
				offset(f, 3);
				break;
			case _fload + 256:
				pushFloat(f, getlocalFloat(f, pc2signedshort(pc)));
				offset(f, 3);
				break;
			case _lload + 256: 
				pushLong(f, getlocalLong(f, pc2signedshort(pc)));
				offset(f, 3);
				break;
			case _dload + 256:
				pushDouble(f, getlocalDouble(f, pc2signedshort(pc)));
				offset(f, 3);
				break;

			case _istore + 256: case _astore + 256:
				setlocal(f, pc2signedshort(pc), pop(f));
				offset(f, 3);
				break;
			case _fstore + 256:
				setlocalFloat(f, pc2signedshort(pc), popFloat(f));
				offset(f, 3);
				break;
			case _lstore + 256: 
				setlocalLong(f, pc2signedshort(pc), popLong(f));
				offset(f, 3);
				break;
			case _dstore + 256:
				setlocalDouble(f, pc2signedshort(pc), popDouble(f));
				offset(f, 3);
				break;

			case _ret + 256:
				f->pc = (uchar *)getlocal(f, pc2signedshort(pc));
				break;
	
			case _iinc + 256:
				setlocal(f, pc2signedshort(pc), getlocal(f, pc2signedshort(pc))
							 + pc2signedshort(pc + 2));
				offset(f, 5);
				break;
			
			case _arraylength: {
				Object *array = (Object *)pop(f);
				
				if( !array ){
					throwException(f, "java/lang/NullPointerException");
					break;
				}
				push(f, arrayLength(array));
				offset(f, 1);
				break;
			}

			case _baload: {
				Object *array;

				index = pop(f);
				array = (Object *)pop(f);
				if( !array ){
					throwException(f, "java/lang/NullPointerException");
					break;
				}
				if( index >= array->vars_size ){
					throwException(f, "java/lang/ArrayIndexOutOfBoundsException");
					break;
				}
				push(f, *((char *)array->vars + index));
				offset(f, 1);
				break;
			}
			case _bastore:  {
				int v = pop(f);
				Object *array;

				index = pop(f);
				array = (Object *)pop(f);
				if( !array ){
					throwException(f, "java/lang/NullPointerException");
					break;
				}
				if( index >= array->vars_size ){
					throwException(f, "java/lang/ArrayIndexOutOfBoundsException");
					break;
				}
				*((char *)array->vars + index) = (char)v;
				offset(f, 1);
				break;
			}

			case _caload: {
				Object *array;

				index = pop(f);
				array = (Object *)pop(f);
				if( !array ){
					throwException(f, "java/lang/NullPointerException");
					break;
				}
				if( index >= (array->vars_size / sizeof(short)) ){
					throwException(f, "java/lang/ArrayIndexOutOfBoundsException");
					break;
				}
				push(f, *((short *)array->vars + index));
				offset(f, 1);
				break;
			}
			case _castore: {
				int v = pop(f);
				Object *array;

				index = pop(f);
				array = (Object *)pop(f);
				if( !array ){
					throwException(f, "java/lang/NullPointerException");
					break;
				}
				if( index >= (array->vars_size / sizeof(short)) ){
					throwException(f, "java/lang/ArrayIndexOutOfBoundsException");
					break;
				}
				*((short *)array->vars + index) = (short)v;
				offset(f, 1);
				break;
			}

			case _aastore: {
				Object *object = (Object *)pop(f);
				Object *array;

				index = pop(f);
				array = (Object *)pop(f);
				if( !array ){
					throwException(f, "java/lang/NullPointerException");
					break;
				}
				if( index >= (array->vars_size / 4) ){
					throwException(f, "java/lang/ArrayIndexOutOfBoundsException");
					break;
				}
				*(array->vars + index) = (int)object;
				offset(f, 1);
				break;
			}
			case _aaload: {
				Object *array;

				index = pop(f);
				array = (Object *)pop(f);
				if( !array ){
					throwException(f, "java/lang/NullPointerException");
					break;
				}
				if( index >= (array->vars_size / 4) ){
					throwException(f, "java/lang/ArrayIndexOutOfBoundsException");
					break;
				}
				push(f, *(array->vars + index));
				offset(f, 1);
				break;
			} 

			case _fastore: {
				float v = popFloat(f);
				Object *array;

				index = pop(f);
				array = (Object *)pop(f);
				if( !array ){
					throwException(f, "java/lang/NullPointerException");
					break;
				}
				if( index >= (array->vars_size / 4) ){
					throwException(f, "java/lang/ArrayIndexOutOfBoundsException");
					break;
				}
				*((float *)array->vars + index) = v;
				offset(f, 1);
				break;
			} 
			case _faload: {
				Object *array;

				index = pop(f);
				array = (Object *)pop(f);
				if( !array ){
					throwException(f, "java/lang/NullPointerException");
					break;
				}
				if( index >= (array->vars_size / 4) ){
					throwException(f, "java/lang/ArrayIndexOutOfBoundsException");
					break;
				}
				pushFloat(f, *((float *)array->vars + index));
				offset(f, 1);
				break;
			}

			case _iaload: {
				Object *array;

				index = pop(f);
				array = (Object *)pop(f);
				if( !array ){
					throwException(f, "java/lang/NullPointerException");
					break;
				}
				if( index >= (array->vars_size / 4) ){
					throwException(f, "java/lang/ArrayIndexOutOfBoundsException");
					break;
				}
				push(f, *(array->vars + index));
				offset(f, 1);
				break;
			}
			case _iastore: {
				int v = pop(f);
				Object *array;

				index = pop(f);
				array = (Object *)pop(f);
				if( !array ){
					throwException(f, "java/lang/NullPointerException");
					break;
				}
				if( index >= (array->vars_size / 4) ){
					throwException(f, "java/lang/ArrayIndexOutOfBoundsException");
					break;
				}
				*(array->vars + index) = v;
				offset(f, 1);
				break;
			}




			case _saload: {
				Object *array;

				index = pop(f);
				array = (Object *)pop(f);
				if( !array ){
					throwException(f, "java/lang/NullPointerException");
					break;
				}
				if( index >= (array->vars_size / 4) ){
					throwException(f, "java/lang/ArrayIndexOutOfBoundsException");
					break;
				}
				push(f, *(array->vars + index));
				offset(f, 1);
				break;
			}
			case _sastore:  {
				int v = pop(f);
				Object *array;

				index = pop(f);
				array = (Object *)pop(f);
				if( !array ){
					throwException(f, "java/lang/NullPointerException");
					break;
				}
				if( index >= (array->vars_size / 4) ){
					throwException(f, "java/lang/ArrayIndexOutOfBoundsException");
					break;
				}
				*(array->vars + index) = v;
				offset(f, 1);
				break;
			}


			case _laload: {
				Object *array;

				index = pop(f);
				array = (Object *)pop(f);
				if( !array ){
					throwException(f, "java/lang/NullPointerException");
					break;
				}
				if( index >= (array->vars_size / 8) ){
					throwException(f, "java/lang/ArrayIndexOutOfBoundsException");
					break;
				}
				pushLong(f, *((vlong *)array->vars + index));
				offset(f, 1);
				break;
			}
			case _lastore: {
				vlong v = popLong(f);
				Object *array;

				index = pop(f);
				array = (Object *)pop(f);
				if( !array ){
					throwException(f, "java/lang/NullPointerException");
					break;
				}
				if( index >= (array->vars_size / 8) ){
					throwException(f, "java/lang/ArrayIndexOutOfBoundsException");
					break;
				}
				*((vlong *)array->vars + index) = v;
				offset(f, 1);
				break;
			}

			case _daload: {
				Object *array;

				index = pop(f);
				array = (Object *)pop(f);
				if( !array ){
					throwException(f, "java/lang/NullPointerException");
					break;
				}
				if( index >= (array->vars_size / 8) ){
					throwException(f, "java/lang/ArrayIndexOutOfBoundsException");
					break;
				}
				pushDouble(f, *((double *)array->vars + index));
				offset(f, 1);
				break;
			} 
			case _dastore:  {
				double v = popDouble(f);
				Object *array;

				index = pop(f);
				array = (Object *)pop(f);
				if( !array ){
					throwException(f, "java/lang/NullPointerException");
					break;
				}
				if( index >= (array->vars_size / 8) ){
					throwException(f, "java/lang/ArrayIndexOutOfBoundsException");
					break;
				}
				*((double *)array->vars + index) = v;
				offset(f, 1);
				break;
			}


			case _nop:
				offset(f, 1);
				break;
			case _pop:
				singlePop(f);
				offset(f, 1);
				break;
			case _pop2:
				nPop(f, 2);
				offset(f, 1);
				break;
			case _dup: {
				int v = pop(f);
				push(f, v); push(f, v);
				offset(f, 1);
				break;
			}
			case _dup2: {
				int v1 = pop(f), v2 = pop(f);
				push(f, v2); push(f, v1); 
				push(f, v2); push(f, v1);
				offset(f, 1);
				break;
			}
			case _dup_x1: {
				int v1 = pop(f), v2 = pop(f);
				push(f, v1); push(f, v2); push(f, v1);
				offset(f, 1);
				break;
			}
			case _dup2_x1: {
				int v1 = pop(f), v2 = pop(f), v3 = pop(f);
				push(f, v2); push(f, v1);
				push(f, v3);
				push(f, v2); push(f, v1);
				offset(f, 1);
				break;
			}
			case _dup_x2: {
				int v1 = pop(f), v2 = pop(f), v3 = pop(f);
				push(f, v1); push(f, v3); 
				push(f, v2); push(f, v1); 
				offset(f, 1);
				break;
			}
			case _dup2_x2: {
				int v1 = pop(f), v2 = pop(f), v3 = pop(f), v4= pop(f);
				push(f, v2); push(f, v1); 
				push(f, v4); push(f, v3); 
				push(f, v2); push(f, v1); 
				offset(f, 1);
				break;
			}
			case _swap: {
				int v1 = pop(f), v2 = pop(f);
				push(f, v1);
				push(f, v2);
				offset(f, 1);
				break;
			}

			case _iadd: 
				push(f, pop(f) + pop(f));
				offset(f, 1);
				break;
			case _ladd: 
				pushLong(f, popLong(f) + popLong(f));
				offset(f, 1);
				break;
			case _fadd:
				pushFloat(f, popFloat(f) + popFloat(f));
				offset(f, 1);
				break;
			case _dadd:
				pushDouble(f, popDouble(f) + popDouble(f));
				offset(f, 1);
				break;

			case _isub: {
				int v2 = pop(f), v1 = pop(f);
				v1 -= v2;
				push(f, v1);
				offset(f, 1);
				break;
			}
			case _lsub: {
				vlong v2 = popLong(f), v1 = popLong(f);
				v1 -= v2;
				pushLong(f, v1);
				offset(f, 1);
				break;
			}
			case _fsub: {
				float f2 = popFloat(f), f1 = popFloat(f);
				f1 -= f2;
				pushFloat(f, f1);
				offset(f, 1);
				break;
			}
			case _dsub: {
				double f2 = popDouble(f), f1 = popDouble(f);
				f1 -= f2;
				pushDouble(f, f1);
				offset(f, 1);
				break;
			}

			case _imul:
				push(f, pop(f) * pop(f));
				offset(f, 1);
				break;
			case _lmul:
				pushLong(f, popLong(f) * popLong(f));
				offset(f, 1);
				break;
			case _fmul: 
				pushFloat(f, popFloat(f) * popFloat(f));
				offset(f, 1);
				break;
			case _dmul: 				
				pushDouble(f, popDouble(f) * popDouble(f));
				offset(f, 1);
				break;

			case _idiv: {
				int v2 = pop(f), v1 = pop(f);
				if( !v2 ){
					throwException(f, "java/lang/ArithmeticException");
					break;
				}
				push(f, v1/v2);
				offset(f, 1);
				break;
			}
			case _ldiv: {
				vlong v2 = popLong(f), v1 = popLong(f);
				if( !v2 ){
					throwException(f, "java/lang/ArithmeticException");
					break;
				}
				pushLong(f, v1 / v2);
				offset(f, 1);
				break;
			}
			case _fdiv: {
				float v2 = popFloat(f), v1 = popFloat(f);
				if( v2 == 0.0 ){
					throwException(f, "java/lang/ArithmeticException");
					break;
				}
				pushFloat(f, v1 / v2);
				offset(f, 1);
				break;
			}
			case _ddiv: {
				double v2 = popDouble(f), v1 = popDouble(f);
				if( v2 == 0.0 ){
					throwException(f, "java/lang/ArithmeticException");
					break;
				}
				pushDouble(f, v1 / v2);
				offset(f, 1);
				break;
			}
			
			case _irem: {
				int v2 = pop(f), v1 = pop(f);
				if( !v2 ){
					throwException(f, "java/lang/ArithmeticException");
					break;
				}
				push(f, v1 % v2);
				offset(f, 1);
				break;
			}
			case _lrem: {
				vlong v2 = popLong(f), v1 = popLong(f);
				if( !v2 ){
					throwException(f, "java/lang/ArithmeticException");
					break;
				}
				pushLong(f, v1 % v2);
				offset(f, 1);
				break;
			}
			case _frem: {
				float v2 = popFloat(f), v1 = popFloat(f);
				if( !v2 ){
					throwException(f, "java/lang/ArithmeticException");
					break;
				}
				pushFloat(f, (float)fmod(v1,v2));
				offset(f, 1);
				break;
			}
			case _drem:{
				double v2 = popDouble(f), v1 = popDouble(f);
				if( v2 == 0.0 ){
					throwException(f, "java/lang/ArithmeticException");
					break;
				}
				pushDouble(f, fmod(v1,v2));
				offset(f, 1);
				break;
			}

			case _ineg:{
				int v1 = pop(f);
				push(f, v1 * -1);
				offset(f, 1);
				break;
			}
			case _lneg: {
				vlong l = popLong(f);
				l = l * -1;
				pushLong(f, l);
				offset(f, 1);
				break;
			}
			case _fneg: 
				pushFloat(f, popFloat(f) * -1.0);
				offset(f, 1);
				break;
			case _dneg: 
				pushDouble(f, popDouble(f) * -1.0);
				offset(f, 1);
				break;
				

			case _ishl: {
				int v2 = pop(f), v1 = pop(f);
				v2 &= 0x1f;
				v1 <<= v2;
				push(f, v1);
				offset(f, 1);
				break;
			}
			case _ishr: {
				int v2 = pop(f), v1 = pop(f);
				v2 &= 0x1f;
				v1 >>= v2;
				push(f, v1);
				offset(f, 1);
				break;
			}
			case _iushr: {
				int v2 = pop(f), v1 = pop(f);
				v2 &= 0x1f;
				if( v1 >= 0 )
					v1 >>= v2;
				else
					v1 = (v1 >> v2) + (2 << (~v2));	
				push(f, v1);
				offset(f, 1);
				break;
			}


			case _lshl: {
				int s = pop(f);
				vlong v = popLong(f);;
				
				v <<= s&0x3f;
				pushLong(f, v);
				offset(f, 1);
				break;
			}
			case _lshr: {
				int s = pop(f);
				vlong v = popLong(f);
				
				v >>= s&0x3f;
				pushLong(f, v);
				offset(f, 1);
				break;
			}
			case _lushr: {
				int s = pop(f);
				vlong v = popLong(f);

				lushr(&v, s);
				pushLong(f, v);
				offset(f, 1);
				break;
			}

			case _iand:
				push(f, pop(f) & pop(f));
				offset(f, 1);
				break;
			case _land: 
				pushLong(f, popLong(f) & popLong(f));
				offset(f, 1);
				break;
			case _ior:
				push(f, pop(f) | pop(f));
				offset(f, 1);
				break;
			case _lor:
				pushLong(f, popLong(f) | popLong(f));
				offset(f, 1);
				break;
			case _ixor:
				push(f, pop(f) ^ pop(f));
				offset(f, 1);
				break;
			case _lxor: 
				pushLong(f, popLong(f) ^ popLong(f));
				offset(f, 1);
				break;

			case _i2l: {
				vlong v = (vlong)pop(f);
				pushLong(f, v);
				offset(f, 1);
				break;
			}
			case _i2f: {
				float v = (float)pop(f);
				pushFloat(f, v);
				offset(f, 1);
				break;
			}
			case _i2d: {
				double v = (double)pop(f);
				pushDouble(f, v);
				offset(f, 1);
				break;
			}
			case _int2byte: {
				uchar c= (uchar)pop(f);
				push(f, (uint)c);
				offset(f,1);
				break;
			}
			case _int2char: {
				ushort s= (ushort)pop(f);
				push(f, (uint)s);
				offset(f,1);
				break;
			}
			case _l2d:
				pushDouble(f, (double)popLong(f));
				offset(f,1);
				break;
			case _d2i:
				push(f, (int)popDouble(f));
				offset(f,1);
				break;
			case _l2i: 
				push(f, (int)popLong(f));
				offset(f,1);
				break;
			case _f2d: 
				pushDouble(f, (double)popFloat(f));
				offset(f,1);
				break;
			case _f2i:
				push(f, (int)popFloat(f));
				offset(f,1);
				break;	
			case _d2l: 
				pushLong(f, (vlong)popDouble(f));
				offset(f,1);
				break;	
			case _d2f:
				pushFloat(f, (float)popDouble(f));
				offset(f,1);
				break;
			case _int2short: { /* A verifier !!!! */
				int v = pop(f);
				short s = (short)v;
			
				push(f, s);
				offset(f, 1);
				break;
			}
			case _f2l:
				pushLong(f, (vlong)popFloat(f));
				offset(f, 1);
				break;
			case _l2f: 
				pushFloat(f, (float)popLong(f));
				offset(f, 1);
				break;  			

			case _ifeq: {
				int v = pop(f);
				if( !v )
					offset(f, pc2signedshort(pc));
				else
					offset(f, 3);
				break;
			}
			case _iflt: {
				int v = pop(f);
				if( v < 0 )
					offset(f, pc2signedshort(pc));
				else
					offset(f, 3);
				break;
			}
			case _ifle: {
				int v = pop(f);
				if( v <= 0 )
					offset(f, pc2signedshort(pc));
				else
					offset(f, 3);
				break;
			}
			case _ifne:{
				int v = pop(f);
				if( v )
					offset(f, pc2signedshort(pc));
				else
					offset(f, 3);
				break;
			}
			case _ifgt: {
				int v = pop(f);
				if( v > 0 )
					offset(f, pc2signedshort(pc));
				else
					offset(f, 3);
				break;
			}
			case _ifge: {
				int v = pop(f);
				if( v >= 0 )
					offset(f, pc2signedshort(pc));
				else
					offset(f, 3);
				break;
			}
			case _if_icmpeq: {
				int v2 = pop(f), v1 = pop(f);
				if( v1 == v2 )
					offset(f, pc2signedshort(pc));
				else
					offset(f, 3);
				break;
			}
			case _if_icmpne: {
				int v2 = pop(f), v1 = pop(f);
				if( v1 != v2 )
					offset(f, pc2signedshort(pc));
				else
					offset(f, 3);
				break;
			}
			case _if_icmplt: {
				int v2 = pop(f), v1 = pop(f);
				if( v1 < v2 )
					offset(f, pc2signedshort(pc));
				else
					offset(f, 3);
				break;
			}
			case _if_icmpgt: {
				int v2 = pop(f), v1 = pop(f);
				if( v1 > v2 )
					offset(f, pc2signedshort(pc));
				else
					offset(f, 3);
				break;
			}
			case _if_icmple: {
				int v2 = pop(f), v1 = pop(f);
				if( v1 <= v2 )
					offset(f, pc2signedshort(pc));
				else
					offset(f, 3);
				break;
			}
			case _if_icmpge: {
				int v2 = pop(f), v1 = pop(f);
				if( v1 >= v2 )
					offset(f, pc2signedshort(pc));
				else
					offset(f, 3);
				break;
			}
			case _lcmp: {
				vlong v2 = popLong(f), v1 = popLong(f);
				if( v1 > v2 )
					push(f, 1LL);
				else if( v1 < v2 )
					push(f, -1LL);
				else
					push(f, 0);
				offset(f, 1);
				break;
			}
			case _fcmpl: {
				float v2 = popFloat(f), v1 = popFloat(f);
				
				if( isNaN(v2) || isNaN(v1) ){
					push(f, -1);
					offset(f, 1);
					break;
				}

				if( v1 > v2 )
					pushFloat(f, 1);
				else if( v1 < v2 )
					push(f, -1);
				else
					push(f, 0);
				offset(f, 1);
				break;
			}
			
			case _fcmpg: {
				float v2 = popFloat(f), v1 = popFloat(f);

				if( isNaN(v2) || isNaN(v1) ){
					push(f, 1);
					offset(f, 1);
					break;
				}

				if( v1 > v2 )
					pushFloat(f, 1);
				else if( v1 < v2 )
					push(f, -1);
				else
					push(f, 0);
				offset(f, 1);
				break;
			}
			case _dcmpl: {
				double v2 = popDouble(f), v1 = popDouble(f);
				
				if( isNaN(v2) || isNaN(v1) ){
					push(f, -1);
					offset(f, 1);
					break;
				}

				if( v1 > v2 )
					push(f, 1);
				else if( v1 < v2 )
					push(f, -1);
				else
					push(f, 0);
				offset(f, 1);
				break;
			}
			case _dcmpg: {
				double v2 = popDouble(f), v1 = popDouble(f);

				if( isNaN(v2) || isNaN(v1) ){
					push(f, 1);
					offset(f, 1);
					break;
				}

				if( v1 > v2 )
					pushFloat(f, 1);
				else if( v1 < v2 )
					push(f, -1);
				else
					push(f, 0);
				offset(f, 1);
				break;
			}


			case _ifnonnull: {				
				if( pop(f) )
					offset(f, pc2signedshort(pc));
				else
					offset(f, 3);
				break;
			}
			case _ifnull: {
				if( !pop(f) )
					offset(f, pc2signedshort(pc));
				else
					offset(f, 3);
				break;
			}


			case _if_acmpeq: {
				if( pop(f) == pop(f) )
					offset(f, pc2signedshort(pc));
				else
					offset(f, 3);
				break;
			}

			case _if_acmpne: {
				if( pop(f) != pop(f) )
					offset(f, pc2signedshort(pc));
				else
					offset(f, 3);
				break;
			}

			case _jsr:
				push(f, (int)(pc + 3));
			case _goto:
				offset(f, pc2signedshort(pc));
				break;

			case _ret:
				f->pc = (uchar *)getlocal(f, pc[1]);
				break;
				
			
			case _return:
				if( f->mb->access_flag&ACC_SYNCHRONIZED )
					releaseLock(f->obj_lock);
DBG1(			prof--;)				
				if( f->invoker ){
					freeFrame(f);
					break;					
				}
				return;
			case _areturn:
			case _ireturn:
				if( f->mb->access_flag&ACC_SYNCHRONIZED )
					releaseLock(f->obj_lock);

DBG1(			prof--;)				
				if( f->invoker ){
					push(f->invoker, pop(f));
					freeFrame(f);
					break;					
				}
				return;
			case _lreturn: 
				if( f->mb->access_flag&ACC_SYNCHRONIZED )
					releaseLock(f->obj_lock);

DBG1(			prof--;)				
				if( f->invoker ){
					pushLong(f->invoker, popLong(f));
					freeFrame(f);
					break;					
				}
				return;
			case _freturn: 
				if( f->mb->access_flag&ACC_SYNCHRONIZED )
					releaseLock(f->obj_lock);

DBG1(			prof--;)				
				if( f->invoker ){
					pushFloat(f->invoker, popFloat(f));
					freeFrame(f);
					break;					
				}
				return;
			case _dreturn: 
				if( f->mb->access_flag&ACC_SYNCHRONIZED )
					releaseLock(f->obj_lock);

DBG1(			prof--;)				
				if( f->invoker ){
					pushDouble(f->invoker, popDouble(f));
					freeFrame(f);
					break;					
				}
				return;

			case _tableswitch: {
				int off =  f->pc - FRAME_CA(f)->code + 1;
				int def, low, hig;
				uchar *n_pc = f->pc;
				
				index = pop(f);

				n_pc += ((off+3)&~3) - off; /* padding */

				def = pc2signedlong(n_pc);
				//n_pc += 4;
				low = pc2signedlong(n_pc + 4);
				//n_pc += 4;
				hig = pc2signedlong(n_pc + 8);
				n_pc += 12;
				/*if( strstr(f->name, "getClassDef") )
					print("%s : def = %d :: low = %d :: hig = %d index = %d\n", 
						f->name, def, low, hig, index);*/
				if( index < low || index > hig ){
					f->pc += def;
				}else{
					n_pc += (index - low) * 4;
					f->pc += pc2signedlong(n_pc);
				}
				
				break;
			}

			case _lookupswitch: {
				int off =  f->pc - FRAME_CA(f)->code + 1;
				int def, npair;
				uchar *n_pc = f->pc;
				int key = pop(f);
			
				n_pc += ((off+3)&~3) - off; /* padding */

				def = pc2signedlong(n_pc);
				n_pc += 4;
				npair = pc2signedlong(n_pc);
				n_pc += 4;			
	
				while( npair-- > 0 ){
					if( pc2signedlong(n_pc) == key ){
						n_pc += 4;			
						f->pc += pc2signedlong(n_pc);
						break;
					}
					n_pc += 8;
				}
				if( npair < 0 )
					f->pc += def;
				
				break;
			}


			case _getstatic: {
				int size;
				char sign[256];

				index = pc2signedshort(pc);
				if( !f->class->r_resolved[index] ){
					cp_resolution(f->class, index, 0);
					TEST_EXCEPTION();
				}
				
				
				{ /* doit disparaite ou introduire _gestatic_qwick */
					
					getSignatureFromCP(sign, index, f->class);
					size = signatureSize(sign);
				}
				
				if( size == WORD ){
					push(f, *((int *)f->class->r_resolved[index]));
				}else if( size == DWORD ){
					pushLong(f, *((vlong *)f->class->r_resolved[index]));
				}else
					panic("Bad size for _getstatic\n");
			
				offset(f, 3);
				break;
			}	
			case _putstatic: {
				int size;
				char sign[256];

				index = pc2signedshort(pc);
				if( !f->class->r_resolved[index] ){
					cp_resolution(f->class, index, 0);
					TEST_EXCEPTION();
				}

				{ /* doit disparaite ou introduire _putstatic_qwick */
					
					getSignatureFromCP(sign, index, f->class);
					size = signatureSize(sign);
				}

				if( size == WORD )
					*((int *)f->class->r_resolved[index]) = pop(f);
				else if( size == DWORD )
					*((vlong *)f->class->r_resolved[index]) = popLong(f);
				else
					panic("Bad size for _putstatic\n");

				offset(f, 3);
				break;
			}
			case _putfield: {
				Object *object;
				char sign[256];
				int size;

				index = pc2signedshort(pc);
				{ /* doit disparaite ou introduire _putfield_qwick */
					
					getSignatureFromCP(sign, index, f->class);
					size = signatureSize(sign);
				}

				if( size == WORD ){
					int i = pop(f);
					object = (Object *)pop(f);
					cp_resolution(f->class, index, object); /* on force la resolution */
					TEST_EXCEPTION();						/* ca doit changer !!!!!! */
					object->vars[f->class->r_resolved[index]-1] = i;
				}else if( size == DWORD ){
					vlong l = popLong(f);
					object = (Object *)pop(f);
					cp_resolution(f->class, index, object); /* on force la resolution */
					TEST_EXCEPTION();						/* ca doit changer !!!!!! */
					*((vlong *)&object->vars[f->class->r_resolved[index]-1]) = l;
				}else
					panic("Bad size for _putfield\n");

				offset(f, 3);
				break;
			}
			case _getfield: {
				Object *object;
				char sign[256];
				int size;
				
				index = pc2signedshort(pc);
				object = (Object *)pop(f);
				///if( !f->class->r_resolved[index] ){
				cp_resolution(f->class, index, object); /* on force la resolution */
				TEST_EXCEPTION();						/* ca doit changer !!!!!! */
				//}
				{ /* doit disparaite ou introduire _getfield_qwick */

					getSignatureFromCP(sign, index, f->class);
					size = signatureSize(sign);
				}

				if( size == WORD ){
					push(f, (object->vars[f->class->r_resolved[index]-1]));
				}else if( size == DWORD ){
					pushLong(f, *((vlong *)&object->vars[f->class->r_resolved[index]-1]));
				}else{
					panic("Bad size for _getfield\n");
				}
					
				offset(f, 3);
				break;
			}

			
			case _invokestatic:{
				Frame *new_frame;
				MethodBlock *mb;
				int i;
				

				index = (f->pc[1]<<8) | f->pc[2];

#ifdef DBG_1
				{ char name[256], sign[256], class_name[256];
					getClassNameFromCP(class_name, index, f->class);
					getNameFromCP(name, index, f->class);
					getSignatureFromCP(sign, index, f->class);
					debug(++prof, "Invokestatic #%d %s.%s%s in class %s\n", 
							index, class_name, name, sign, f->class->r_name);
				} 
#endif
				if( !f->class->r_resolved[index] ){
					cp_resolution(f->class, index, 0);
					TEST_EXCEPTION();
				}
	
				mb = (MethodBlock *)f->class->r_resolved[index];

				if( mb->access_flag&ACC_ABSTRACT ){ /* si methode abstraite */
					throwException(f, "java/lang/AbstractMethodError");
					break;
				}

				if( mb->access_flag&ACC_NATIVE ){
					nativeCall(mb, 3, 1);
					break;
				}

				if( (mb->access_flag&ACC_SYNCHRONIZED && 
						getLock(mb->class->r_java_class) > 0) ||
						!(mb->access_flag&ACC_SYNCHRONIZED) ){

					offset(f, 3);
					new_frame =  newFrame(mb, f);
					if( !new_frame )
						panic("can't create frame Invokevirtual()");
					
					for(i=0 ; i < mb->nargs ; i++)
						setlocal(new_frame, mb->nargs-i-1, pop(f->invoker));
						
					if( mb->access_flag&ACC_SYNCHRONIZED )
						f->obj_lock = mb->class->r_java_class;
				}

				break;
			}
			case _invokenonvirtual: { /* == invokespecial */
				Frame *new_frame;
				MethodBlock *mb;
				int i;

				index = (f->pc[1]<<8) | f->pc[2];

#ifdef DBG_1
				{ char name[256], sign[256], class_name[256];
					getClassNameFromCP(class_name, index, f->class);
					getNameFromCP(name, index, f->class);
					getSignatureFromCP(sign, index, f->class);
					debug(++prof, "Invokenonvirtual #%d %s.%s%s in class %s\n", 
							index, class_name, name, sign, f->class->r_name);
				} 
#endif
				if( !f->class->r_resolved[index] ){
					cp_resolution(f->class, index, 0);
					TEST_EXCEPTION();
				}
				
				mb = (MethodBlock *)f->class->r_resolved[index];

				if( mb->access_flag&ACC_ABSTRACT ){ /* si methode abstraite */
					throwException(f, "java/lang/AbstractMethodError");
					break;
				}

				if( mb->access_flag&ACC_NATIVE ){
					nativeCall(mb, 3, 0);
					break;
				}

				if( (mb->access_flag&ACC_SYNCHRONIZED && 
						getLock((Object *)f->opstack[f->opstack_pt - mb->nargs]) > 0) ||
						!(mb->access_flag&ACC_SYNCHRONIZED) ){

					offset(f, 3);
					new_frame =  newFrame(mb, f);
					if( !new_frame )
						panic("can't create frame Invokevirtual()");
					
					/* passage d'arguments */
					for(i=mb->nargs ; i >= 0  ; i--)
						setlocal(new_frame, i, pop(f->invoker));
						
					if( mb->access_flag&ACC_SYNCHRONIZED )
						f->obj_lock = (Object *)new_frame->locals[0];
				}


				break;
			}
			case _invokevirtual: {
				ClassFile *cur_class;
				Frame *new_frame;
				MethodBlock *mb;
				Object *obj;
				int i, r, off, args;

				cur_class = f->class;
				index = (f->pc[1]<<8) | f->pc[2];

#ifdef DBG_1
				{ char name[256], sign[256], class_name[256];
					getClassNameFromCP(class_name, index, f->class);
					getNameFromCP(name, index, f->class);
					getSignatureFromCP(sign, index, f->class);
					debug(++prof, "Invokenonvirtual #%d %s.%s%s in class %s\n", 
							index, class_name, name, sign, f->class->r_name);
				} 
#endif

				if( !cur_class->r_resolved[index] )
					cp_resolution(cur_class, index, (Object *)1);
					TEST_EXCEPTION();

				r = cur_class->r_resolved[index];
				off = r >> 16;
				args = r&0xFFFF;
				obj = (Object *)f->opstack[f->opstack_pt - args];
				if( !obj ){
					throwException(f, "java/lang/NullPointerException");
					break;
				}
				mb = obj->class->r_methods + off;
				
				if( mb->access_flag&ACC_ABSTRACT ){ /* si methode abstraite */
					throwException(f, "java/lang/AbstractMethodError");
					break;
				}

				if( mb->access_flag&ACC_NATIVE ){
					nativeCall(mb, 3, 0);
					break;
				}

				if( (mb->access_flag&ACC_SYNCHRONIZED && getLock(obj) > 0) ||
						!(mb->access_flag&ACC_SYNCHRONIZED) ){

					offset(f, 3);
					new_frame =  newFrame(mb, f);
					if( !new_frame )
						panic("can't create frame Invokevirtual()");
					
					/* passage d'arguments */
					for(i=mb->nargs ; i >= 0  ; i--)
						setlocal(new_frame, i, pop(f->invoker));
						
					if( mb->access_flag&ACC_SYNCHRONIZED )
						f->obj_lock = obj;
				}

				break;
			}

			case _invokeinterface: {
				ClassFile *class_obj;
				Frame *new_frame;
				MethodBlock *mb;
				Object *obj;
				char name[256], sign[256], class_name[256], buf[256];
				int i, nargs, n;

				index = (f->pc[1]<<8) | f->pc[2];
				nargs = f->pc[3];

				getClassNameFromCP(class_name, index, f->class);
				getNameFromCP(name, index, f->class);
				getSignatureFromCP(sign, index, f->class);

DBG1(			debug(++prof,"Invokeinterface #%d %s.%s%s in class %s\n", 
						index, class_name, name, sign, f->class->r_name); )

				obj = (Object *)f->opstack[f->opstack_pt - nargs +1];

				class_obj = obj->class;
				mb = class_obj->r_methods;

				sprint(buf, "%s%s", name, sign);

				//for (n=0; n < class_obj->r_n_methods ; n++)
				for (n=class_obj->r_n_methods-1; n >= 0 ; n--)
					if( !strcmp(buf, mb[n].name) )
						break;
				if( n < 0 )
					panic("can't find method block for %s in invokeinterface\n", buf);
				
				mb = mb + n;

				if( mb->access_flag&ACC_ABSTRACT ){ /* si methode abstraite */
					throwException(f, "java/lang/AbstractMethodError");
					break;
				}

				if( mb->access_flag&ACC_SYNCHRONIZED )
					panic("synchronised method in _invokeinterface\n");

				if( mb->access_flag&ACC_NATIVE ){ /* execution d'une méthode native */
					int *tmp; char buf[256];

					sprint(buf, "%s.%s", class_obj->r_name, name);
					tmp = (int *)sysMalloc((mb->nargs + 1)*sizeof(int));
					for(i=mb->nargs ; i >= 0  ; i--)
						tmp[i] = pop(f);

					offset(f, 5);

					//invokeNative(f, buf, sign, (arg_list)tmp);
					panic("invokeInterface as native()\n");

					free(tmp);
DBG1(				prof--; )
					break;
				}

				offset(f, 5);

				/* Creation de la nouvelle frame */
	
				new_frame =  newFrame(mb, f);
				if( !new_frame )
					panic("can't create frame Invokevirtual()");

				/* passage d'argument */
				for(i=mb->nargs ; i >= 0  ; i--)
					setlocal(new_frame, i, pop(f->invoker));

				/* exécution */
				break;
			}


			case _new: {
				ClassFile *new_class;
				Object *new_object;

				index = pc2signedshort(pc);
				if( !f->class->r_resolved[index] ){
					cp_resolution(f->class, index, 0);
					TEST_EXCEPTION();
DBG1(				if( !f->class->r_resolved[index] ) 
						panic("cp[%d] not resolved new\n", index); )
				}

				new_class = (ClassFile *)f->class->r_resolved[index];
				new_object = newObject(new_class);
				push(f, (int)new_object);
				offset(f, 3);

				break;
			}
			
			case _newarray: {
				Object *new_array;
				int v = pop(f);

				if( v < 0 ){
					throwException(f, "java/lang/NegativeArraySizeException");
					break;
				}
				new_array = newArray((int)pc[1], v);
				push(f, (int)new_array);
				offset(f, 2);

				break;
			}
			case _anewarray: {
				Object *new_array;
				int v = pop(f);
				
				index = pc2signedshort(pc);
				if( !f->class->r_resolved[index] ){
					cp_resolution(f->class, index, 0);
					TEST_EXCEPTION();
				}
				if( f->class->r_resolved[index] > 0 ){ /* classe */
					new_array = 
						newObjectArray((ClassFile *)f->class->r_resolved[index], v);
				}else{	/* primitive type */
					new_array = newObjectArray((ClassFile *)1, v);
				}
				push(f, (int)new_array);
				offset(f, 3);
				break;
			}

			case _multianewarray: {
				Object *new_array;
				int dimension = pc[3];
				int i, *vdim;
				
				index = pc2signedshort(pc);
				if( dimension < 1 )
					panic("dimmension must be >= 1 multianewarray\n");

				vdim = (int *) sysMalloc(dimension*sizeof(int));
				if( !vdim )
					panic("malloc failed multianewarray\n");
			
				for(i=0 ; i < dimension ; i++)
					if( (vdim[dimension-i-1] = pop(f)) < 0 ){
						throwException(f, "java/lang/NegativeArraySizeException");
						free(vdim);
						break;
					}
				if( i < dimension )
					break;	 
				new_array = newMultiArray(f->class, index, dimension, vdim);
				push(f, (int)new_array);
				free(vdim);
				offset(f, 4);
				break;
			}

			/*case _newfromname: */
			/*case _verifystack:*/
			case _checkcast: {
				Object * obj = (Object *)pop(f);
				int ans = 0;
				ClassFile *c; //char buf[256];

				index = pc2signedshort(pc);
				if( !obj ){
					push(f, (int)obj);
					offset(f, 3);
					break;
				}

				if( !f->class->r_resolved[index] ){
					cp_resolution(f->class, index, 0);
					TEST_EXCEPTION();
				}
				c = (ClassFile *)f->class->r_resolved[index];

/*				print("obj %s :: class ", obj->class->r_name);
				getClassFromCP(buf, index, f->class);
				print("%s\n", buf);
				if( IS_OBJECT(obj) ){
					print("1\n");
					ans = isCompatibleObject(obj->class, c);
				}else if( IS_ARRAY(obj) || IS_ARRAY_OBJECT(obj) ){
					print("2\n");
					dumpObject(obj);
					ans = isCompatibleArray(obj, buf);
				}
*/
				if( IS_OBJECT(obj) ){
					if( c->access_flags&ACC_INTERFACE && isImplementOf(obj->class, c)){
						ans = 1;
					}else if( isSubClassOf(obj->class, c) )
						ans = 1;		
				}else if( IS_ARRAY(obj) || IS_ARRAY_OBJECT(obj) ){

					if( !isClassArray(index, f->class) ){
						if( c == getClass("java/lang/Object") )
							ans = 1;
					}else{
						if( (classArrayType(index, f->class) != T_CLASS) &&
								IS_ARRAY(obj) && 
								(obj->type == classArrayType(index, f->class)) )
							ans = 1;
						else if( (classArrayType(index, f->class) == T_CLASS) &&
								IS_ARRAY_OBJECT(obj) /*&&
								isSubClassOf(obj->class, c)*/	){
							//print("I'm not sure of that  !!!!\n");
							ans = 1;
						}
					}
				}
				push(f, (int)obj);
				if( ans < 0 /*!ans ATTENTION !!!!!!!! */ ){
					throwException(f, "java/lang/ClassCastException");
					break;
				}
				offset(f, 3);
				break;
			}

			case _instanceof: {
				Object * obj = (Object *)pop(f);
				int ans = 0;
				ClassFile *c;//char buf[256];

				index = pc2signedshort(pc);
				if( !obj ){
					push(f, 0);
					offset(f, 3);
					break;
				}

				if( !f->class->r_resolved[index] ){
					cp_resolution(f->class, index, 0);
					TEST_EXCEPTION();
				}
				c = (ClassFile *)f->class->r_resolved[index];

/*				getClassFromCP(buf, index, f->class);
				if( IS_OBJECT(obj) ){
					ans = isCompatibleObject(obj->class, c);
				}else if( IS_ARRAY(obj) || IS_ARRAY_OBJECT(obj) ){
					ans = isCompatibleArray(obj, buf);
				}
*/
				if( IS_OBJECT(obj) ){
					if( c->access_flags&ACC_INTERFACE && isImplementOf(obj->class, c)){
						ans = 1;
					}else if( isSubClassOf(obj->class, c) )
						ans = 1;		
				}else if( IS_ARRAY(obj) || IS_ARRAY_OBJECT(obj) ){

					if( !isClassArray(index, f->class) ){
						if( c == getClass("java/lang/Object") )
							ans = 1;
					}else{
						if( (classArrayType(index, f->class) != T_CLASS) &&
								IS_ARRAY(obj) && 
								(obj->type == classArrayType(index, f->class)) )
							ans = 1;
						else if( (classArrayType(index, f->class) == T_CLASS) &&
								IS_ARRAY_OBJECT(obj) /*&&
								isSubClassOf(obj->class, c)*/	){
							//print("I'm not sure of that  !!!!\n");
							ans = 1;
						}
					}
				}
			
				if( ans )
					push(f, 1);
				else 
					push(f, 0);
				offset(f, 3);
				break;
			}

			case _monitorexit: {
				Object *obj = (Object *)pop(f);
				
				//print("exit  : %d ref = %d (%d)\n", obj, obj->s.ref, current_thread);

				releaseLock(obj);

				offset(f, 1);
				break;
			}	
			case _monitorenter:	{
				Object *obj = (Object *)pop(f);
				
				//print("enter : %d ref = %d (%d)\n", obj, obj->s.ref, current_thread);

				if( getLock(obj) < 0 ){ /* on a pas le lock */
					push(f, (int)obj);
					break;
				}
				f->obj_lock = obj;
	
				offset(f, 1);
				break;
			}

			case _breakpoint:
				goto notyet;
			
			internal_nop:		/* ne fait absolument rien */
				offset(f, 0);	/* crée pour continuer sur une exception */
				break;			/* quand elle est attrappée */

			case _athrow: {
				int i, where_pc;
				Frame *finvk;
				Object *obj = (Object *)f->opstack[f->opstack_pt];
					
DBG1(			chat("_athrow :\n");
				chat("I am in %s.%s\n", f->class->r_name, f->name); )

				for(i=0 ; i < FRAME_CA(f)->exception_table_length ; i++){
#ifdef DBG_1
					chat("exception %d :", i);
					chat("from %d to %d at %d ", 
								FRAME_CA(f)->exception_table[i].start_pc,
								FRAME_CA(f)->exception_table[i].end_pc,
								FRAME_CA(f)->exception_table[i].handler_pc);
					if( FRAME_CA(f)->exception_table[i].catch_type ){
						char name[256];
						cp_info *cp = &f->class->constant_pool[FRAME_CA(f)->exception_table[i].catch_type];
					
						GetCpString(name, cp->u.c.name_index, f->class);
						chat("%s \n", name);
					}else{
						chat("all exception \n");
						panic("not yet part implemeted in _athrow\n");
						/* si si c'est géré, c'est juste pour me prévenir quand
						   cela se produira */
					}
#endif
					where_pc = f->pc - FRAME_CA(f)->code ;
					if( where_pc >= FRAME_CA(f)->exception_table[i].start_pc &&
										where_pc < FRAME_CA(f)->exception_table[i].end_pc ){
						char name[256];
						cp_info *cp = &f->class->constant_pool[FRAME_CA(f)->exception_table[i].catch_type];
						ClassFile *super;

						GetCpString(name, cp->u.c.name_index, f->class);
				
						super = lookupClass(name);

						if( !strcmp(name, obj->class->r_name) || isSubClassOf(obj->class, super) ||
								FRAME_CA(f)->exception_table[i].catch_type == 0 ){
							f->pc = FRAME_CA(f)->code + FRAME_CA(f)->exception_table[i].handler_pc;
							f->exception = 0;
							f->opstack_pt = 0; f->opstack[0] = (int)obj;
							goto internal_nop; /* pour continuer l'exécution sur le nouveau pc */
						}

					}	

				}
DBG1(			chat("no exception catch %s\n", obj->class->r_name); )
				if( f->mb->access_flag&ACC_SYNCHRONIZED )
					releaseLock(f->obj_lock);
				finvk = findInvokerFrame(f);
				if( finvk ){
					finvk->opstack[finvk->opstack_pt] = (int)obj;
					finvk->exception = obj;
					f->opstack_pt = -1;
DBG1(				chat("returning to invoker\n\n"); );
					freeFrame(f);
				}else{
DBG1(				chat("returning to invoker\n\n"); );
					f->exception = obj;	
					return;
				}
				break;
			}
			
			default:
				panic("unknown instruction %d %s\n",
						opcode, Instructions[opcode&0xFF].name);
			notyet:
				panic("not implemented instruction %s (%d)\n",
						Instructions[opcode&0xFF].name, opcode&0xFF);
			

		} /* end switch */

#ifdef PROFILE
		inst_prof[opcode&0xFF].time += getMiliSec() - cur_time; 
#endif

	} /* end for */

}

static void 
staticNativeCall(MethodBlock *mb)
{
	char buf[256];
	int *tmp, i;
	ClassFile *cur_class;

	cur_class = mb->class;
	
	if( mb->access_flag&ACC_SYNCHRONIZED ){
		if( getLock(cur_class->r_java_class) < 0 )
				return;
		f->obj_lock = cur_class->r_java_class;
	}

	sprint(buf, "%s.%s", cur_class->r_name, mb->name);

	tmp = (int *)sysMalloc(mb->nargs*sizeof(int));
	if( !tmp )
		panic("sysMalloc failed in nativeCall()\n");
	for(i=0 ; i < mb->nargs ; i++)
		tmp[mb->nargs-i-1] = pop(f);

	offset(f, 3);

	invokeNative(f, buf, (arg_list)tmp);

	if( mb->access_flag&ACC_SYNCHRONIZED )
		releaseLock(f->obj_lock);

	free(tmp);
}
static void 
nCall(MethodBlock *mb)
{
	char buf[256];
	int *tmp, i;
	ClassFile *cur_class;

	cur_class = mb->class;
	
	if( mb->access_flag&ACC_SYNCHRONIZED ){
		if( getLock((Object *)f->opstack[f->opstack_pt - mb->nargs]) < 0 )
			return;
		f->obj_lock = (Object *)f->opstack[f->opstack_pt - mb->nargs];;
	}

	sprint(buf, "%s.%s", cur_class->r_name, mb->name);

	tmp = (int *)sysMalloc((mb->nargs + 1)*sizeof(int));
	if( !tmp )
		panic("sysMalloc failed in nativeCall()\n");

	for(i=mb->nargs ; i >= 0  ; i--)
		tmp[i] = pop(f);

	offset(f, 3);

	invokeNative(f, buf, (arg_list)tmp);
		
	if( mb->access_flag&ACC_SYNCHRONIZED )
		releaseLock(f->obj_lock);

	free(tmp);
}

static void 
nativeCall(MethodBlock *mb, int pc_offset, int is_static)
{
	char buf[256];
	int *tmp, i, nargs;
	ClassFile *cur_class;
	Object *obj;

	cur_class = mb->class;
	if( is_static ){
		nargs = mb->nargs;
		obj = cur_class->r_java_class;
	}else{
		nargs = mb->nargs + 1;
		obj = (Object *)f->opstack[f->opstack_pt - mb->nargs];
	}

	if( mb->access_flag&ACC_SYNCHRONIZED ){
		if( getLock(obj) < 0 )
			return;
		f->obj_lock = obj;
	}

	sprint(buf, "%s.%s", cur_class->r_name, mb->name);
	
	tmp = (int *)sysMalloc(nargs*sizeof(int));
	if( !tmp )
		panic("sysMalloc failed in nativeCall()\n");

	if( is_static )
		for(i=0 ; i < mb->nargs ; i++)
			tmp[mb->nargs-i-1] = pop(f);
	else
		for(i=mb->nargs ; i >= 0  ; i--)
			tmp[i] = pop(f);
	
	offset(f, pc_offset);

	invokeNative(f, buf, (arg_list)tmp);
		
	if( mb->access_flag&ACC_SYNCHRONIZED )
		releaseLock(obj);

	free(tmp);
}

void
executeVirtual(ClassFile *class, Object *object, char *func, char *signature,...)
{
	arg_list args;
	Frame *frame;
	int nargs;
	char *p = signature;	
	
	//print("executeVirtual()\n");

	arg_begin(args, signature);

	/* frame du constructeur <init> ()V */
	frame = newFrameFromName(class, func, signature);
	if( !frame )
		panic("can't create new Frame in executeVirtual() %s %s\n", func, signature);
	
	/* preparation des variables locales */
	frame->locals[0] = (int)object;

	/* construction des arguments */

	while( *p++ != '(' );
	
	for(nargs = 0; *p != ')' && *p ; p++){
		switch( *p ){
			case '[': /* tableau */
				setlocal(frame, nargs + 1, (int)arg_value(args, Object *));
				nargs += 1;
				while( *p == '[' ) p++;
				if( *p == 'L' ) 
					while( *p != ';' ) p++;
				break;
			case 'L': /* classe */
				setlocal(frame, nargs + 1, (int)arg_value(args, Object *));
				nargs += 1;
				while( *p != ';' ) p++;
				break;
			case 'I': /* integer */
			case 'Z': /* boolean */
			case 'C': /* char    */
			case 'S': /* short   */
			case 'B': /* byte	 */
				setlocal(frame, nargs + 1, arg_value(args, int));
				nargs += 1;
				break;
			case 'F': { /* float   */
				float *f = arg_value(args, float *);
				setlocalFloat(frame, nargs + 1, *f);
				nargs += 1;
				break;
			}
			case 'D': /* double */
				setlocalDouble(frame, nargs + 1, arg_value(args, double));
				nargs += 2;
				break;
			case 'J': /* long */
				setlocalLong(frame, nargs + 1, arg_value(args, vlong));
				nargs += 2;
				break;
			default:
				panic("unknown signature %s executeVirtual()\n", signature);
		}
	}
	/* exécution */

	if( frame->mb->access_flag&ACC_SYNCHRONIZED ){
		if( getLock(object) < 0 )
			panic("getLock in executeVirtual()\n");
		frame->obj_lock = object;
	}

	interpreter();
	freeFrame(frame);
}

void 
executeConstructor(Object *object, char *signature,...)
{
	arg_list args;
	Frame *cons_frame;
	int nargs;
	char *p = signature;	
	
	//print("executeConstructor()\n");

	arg_begin(args, signature);

	/* frame du constructeur <init> ()V */
	cons_frame = newFrameFromName(object->class, "<init>", signature);
	if( !cons_frame )
		panic("can't create new Frame in executeConstructor()\n");
	
	/* preparation des variables locales */
	cons_frame->locals[0] = (int)object;

	/* construction des arguments */

	while( *p++ != '(' );
	
	for(nargs = 0; *p != ')' && *p ; p++){
		switch( *p ){
			case '[': /* tableau */
				setlocal(cons_frame, nargs + 1, (int)arg_value(args, Object *));
				nargs += 1;
				while( *p == '[' ) p++;
				if( *p == 'L' ) 
					while( *p != ';' ) p++;
				break;
			case 'L': /* classe */
				setlocal(cons_frame, nargs + 1, (int)arg_value(args, Object *));
				nargs += 1;
				while( *p != ';' ) p++;
				break;
			case 'I': /* integer */
			case 'Z': /* boolean */
			case 'C': /* char    */
			case 'S': /* short   */
			case 'B': /* byte	 */
				setlocal(cons_frame, nargs + 1, arg_value(args, int));
				nargs += 1;
				break;
			case 'F': { /* float   */
				float *f = arg_value(args, float *);
				setlocalFloat(cons_frame, nargs + 1, *f);
				nargs += 1;
				break;
			}
			case 'D': /* double */
				setlocalDouble(cons_frame, nargs + 1, arg_value(args, double));
				nargs += 2;
				break;
			case 'J': /* long */
				setlocalLong(cons_frame, nargs + 1, arg_value(args, vlong));
				nargs += 2;
				break;
			default:
				panic("unknown signature %s executeConstructor()\n", signature);
		}
	}
	/* exécution du constructeur */
	interpreter();
	freeFrame(cons_frame);
}

vlong
popLong(Frame *f)
{
	union { vlong v; uint t[2]; } u;

	u.t[1] = f->opstack[f->opstack_pt--];
	u.t[0] = f->opstack[f->opstack_pt--];

	return u.v;
}
double
popDouble(Frame *f)
{
	union { double v; uint t[2]; } u;

	u.t[1] = f->opstack[f->opstack_pt--];
	u.t[0] = f->opstack[f->opstack_pt--];

	return u.v;

}

#ifdef PROFILE
void dumpProfile(void)
{
	int i;

	for(i=0 ; i < 256 ; i++)
		if( inst_prof[i].nb && Instructions[i].name )
		fprint(2, "%10d %5d %4.3f %s\n", 
				inst_prof[i].nb, inst_prof[i].time, 
				((float)inst_prof[i].time / (float)inst_prof[i].nb),
				Instructions[i].name);

}
#endif



