/* 
 * Déclaration de constantes
 * pour la lecture de fichiers java
 * compilés
 */

typedef unsigned char  u1;
typedef unsigned short u2;
typedef unsigned long  u4;

typedef struct MethodBlock MethodBlock;
typedef struct Object Object;

typedef struct ClassFile ClassFile;

typedef struct cp_info   cp_info;
typedef struct cp_class cp_class;
typedef struct cp_method cp_method;
typedef struct cp_string cp_string;
typedef struct cp_int_float cp_int_float;
typedef struct cp_long_double cp_long_double;
typedef struct cp_nametype cp_nametype;
typedef struct cp_asciz cp_asciz;

typedef struct field_info field_info;
typedef struct method_info method_info;

typedef struct attribute_info attribute_info;
typedef struct Code_attribute Code_attribute;
typedef struct SourceFile_attribute SourceFile_attribute;
typedef struct Exception_attribute Exception_attribute;
typedef struct exception exception;
typedef struct LineNumberTable_attribute LineNumberTable_attribute;
typedef struct LocalVariableTable_attribute LocalVariableTable_attribute;

#define CLASSEFILE_MAGIC    0xCAFEBABE

enum {
	T_CLASS = 0,
	T_TAB,
	T_STRING,
	T_VOID,
	T_BOOLEAN,
	T_CHAR,
	T_FLOAT,
	T_DOUBLE,
	T_BYTE,
	T_SHORT,
	T_INT,
	T_LONG
};

struct ClassFile {
	/* 
	 * Format from file.class 
	 */
	u4 			magic;
	u2 			minor_version;
	u2 			major_version;
  
	u2 			constant_pool_count;
	cp_info *	constant_pool;
  
	u2 			access_flags;
	u2 			this_class;
	u2 			super_class;
  
	u2 			interfaces_count;
	u2 *		interfaces;
  
	u2 			fields_count;
	field_info *fields;
  
	u2 			methods_count;
	method_info *methods;

	u2 			attributes_count;
	attribute_info *attributes;

	/*
	 * runtime infomation 
	 */
	
	char *			r_name;
	char *			r_file;
	int 			r_state;/* CF_LINKED, CF_INITIALIZED */
	ClassFile *		r_super;
	Object *		r_java_class;
	int *			r_fields;
	int *			r_resolved;
	MethodBlock *	r_methods;
	int 			r_n_methods;
};

#define CF_LINKED				1
#define CF_INITIALIZED			2
#define CF_BEING_INITIALIZED	4


/*
 * CONSTANT POOL declarations
 */

enum {
	CONSTANT_Asciz = 1, /* same as Utf_8 */
	CONSTANT_Unicode, 	/* unused */
	CONSTANT_Integer,
	CONSTANT_Float,
	CONSTANT_Long,
	CONSTANT_Double,
	CONSTANT_Class,
	CONSTANT_String,
	CONSTANT_Fieldref,
	CONSTANT_Methodref,
	CONSTANT_InterfaceMethodref,
	CONSTANT_NameAndType
};

struct cp_class { u2 name_index; };
struct cp_method { u2 class_index; u2 name_and_type_index; };
struct cp_string { u2 string_index; };
struct cp_int_float { u4 bytes; };
struct cp_long_double { u4 hight_bytes; u4 low_bytes; };
struct cp_nametype { u2 name_index; u2 signature_index; };
struct cp_asciz { u2 length; u1 *bytes; };

struct cp_info {
	u1 tag;	/* CONSTANT_* */
	union {
		cp_class c;
		cp_method m;
		cp_string s;
		cp_int_float i;
		cp_long_double l;
		cp_nametype n;
		cp_asciz a;
	} u;
};

/* 
 * Fields declaration
 */

struct field_info {
	u2 access_flag;
	u2 name_index;
	u2 signature_index;
	u2 attributes_count;
	attribute_info *attributes;
};

/* Access flags */

#define ACC_PUBLIC			0x0001
#define ACC_PRIVATE 		0x0002
#define ACC_PROTECTED		0x0004
#define ACC_STATIC			0x0008
#define ACC_FINAL			0x0010
#define ACC_SYNCHRONIZED	0x0020
#define ACC_VOLATILE		0x0040
#define ACC_TRANSIENT		0x0080
#define ACC_NATIVE			0x0100
#define ACC_INTERFACE		0x0200
#define ACC_ABSTRACT		0x0400

/* 
 * Methods declaration
 */

#define ATTR_CODE			0x01
#define ATTR_EXCEPTION		0x02
#define ATTR_LINE_NUMBER	0x03
#define ATTR_LOCAL_VARIABLE	0x04

struct method_info {
	u2 access_flag;
	u2 name_index;
	u2 signature_index;
	u2 attributes_count;
	struct method_attribute {
		u1 type;	/*  ATTR_CODE or ATTR_EXCEPTION */
		union {
			Code_attribute *ca;
			Exception_attribute *ea;
		} u;
	} *attributes;
};

/* 
 * Attributes declaration 
 */

struct attribute_info { /* generique attribute */
	u2 attribute_name;
	u4 attribute_length;
	u1 *info;
};

struct LineNumberTable_attribute {
	u2 attribute_name_index;
	u4 attribute_length;
	u2 line_number_table_length;
	struct line_number {
		u2 start_pc;
		u2 line_number;
	} *line_number_table;
};

struct LocalVariableTable_attribute {
	u2 attribute_name_index;
	u4 attribute_length;
	u2 local_variable_table_length;
	struct local_variable {
		u2 start_pc;
		u2 length;
		u2 name_index;
		u2 signature_index;
		u2 slot;
	} *local_variable_table;
};

struct Code_attribute {
	u2 attribute_name_index;
	u4 attribute_length;
	u2 max_stack;
	u2 max_locals;
	u4 code_length;
	u1 *code;
	u2 exception_table_length;
	struct exception {
		u2 start_pc;
		u2 end_pc;
		u2 handler_pc;
		u2 catch_type;
	} *exception_table;
	u2 attribute_count;
	struct code_attribute {
		u1 type;	/* ATTR_LINE_NUMBER or ATTR_LOCAL_VARIABLE */
		union {
			LineNumberTable_attribute la;
			LocalVariableTable_attribute va;
		} u;
	} *attribute;
};

struct Exception_attribute {
	u2 attribute_name_index;	/* not set actually */
	u4 attribute_length;		/* not set actually */
	u2 number_of_exception;
	u2 *exception_index_table;
};

struct SourceFile_attribute {
	u2 attribute_name_index;
	u2 attribute_length;
	u2 source_file_index;
};




