#include <u.h>
#include <libc.h>
#include <lock.h>

#include "java_const.h"
#include "runtime.h"
#include "fns.h"


static void startJVM(char *, char **);
static void dumpStackTrace(Object *);
static void usage(char *av0);

char *java_version		= "0.1";
char *java_vendor	 	= "Bodet Laurent dept. MIME";
char *java_vendor_url 	= "http://www.mime.univ-paris8.fr/~bl";
char *java_home 		= 0; /* get from environement by default */
char *java_class_path 	= 0; /* get from environement by default */
int java_max_stack 		= 16000 * 1024;

extern StackFrame *current_stack;
extern Frame *current_frame;

void
main(int argc, char **argv)
{
	char *av0 = argv[0];

	argv++, argc--;

	while( argc && *argv[0] == '-' ){
		switch((*argv)[1]){
			case 'v':
				if( strcmp((*argv)+1, "version") )
					goto bad_option;
				fprint(2, "version %s\n", java_version);
				exits("ok");
				break;
			case 'c':
				if( strcmp((*argv)+1, "classpath") )
					goto bad_option;
				java_class_path = *++argv; argc--;
				if( !java_class_path ){
					fprint(2, "missing directory list for option -classpath\n");
					usage(av0);
				}
				break;
			case 's':
				if( strcmp((*argv)+1, "stacksize") )
					goto bad_option;
				java_max_stack = atoi(*++argv); argc--;
				if( java_max_stack <= 0 ){
					fprint(2, "missing or bad size for -stacksize\n");
					usage(av0);
				}
				break;
			bad_option:
			default:
				fprint(2, "unknown option %s\n", *argv);
				usage(av0);
		}
		argv++, argc--;
	}


	if( argc && *argv ){ /* on peut lancer l'exécution */
		
		/* trouve le class_path */
		if( !java_class_path ){ 
			java_class_path = getenv("java_class_path");
			if( !java_class_path || !java_class_path[0] )
				java_class_path = ".";
		}

		/* touve java_home pour le compilateur (javac.properties) */
		java_home = getenv("java_home");
		if( !java_home || !java_home[0] )
			java_home = "."; 
		
		makePath();
		startJVM(*argv, argv+1);
#ifdef PROFILE
		dumpProfile();
#endif
		exits("ok");
	}

	usage(av0);
}

static void usage(char *av0)
{
	fprint(2, "usage: %s [-option] class\n\n", av0);
	
	fprint(2, "options currently supported:\n");
	fprint(2, "-version\t\tprint out the build version.\n");
	fprint(2, "-classpath <directories separated by spaces>\n"
				"\t\t\tlist directories in which to look for classes.\n"
				"\t\t\tex. -classpath \'. /lib/java/classes.zip\'\n");
	fprint(2, "-stacksize\t\tmax size for the running stack\n");

	exits("usage");
}



/* Start JVM */

static void
startJVM(char *class_name, char **av)
{
	ClassFile *main_class, *string_class;
	Frame *main_frame;
	
	{	/* load java Class ==> load Object too*/
		ClassFile *c;

		c = lookupClass("java/lang/Class");
		if( !c )
			panic("class java.lang.Class not found\n");
	}
	initThread();
	//current_stack = initGlobalStack(0);

DBG1(chat("\nBegin loading, linking...:\n");)

	point2slash(class_name);
	main_class = lookupClass(class_name);
	if( !main_class )
		panic("class %s not found\n", class_name);

	string_class = lookupClass("java/lang/String"); 		
	if( !string_class )
		panic("class class java.lang.String not found\n");		

	main_frame = newFrameFromName(main_class, "main", "([Ljava/lang/String;)V");
	if( !main_frame )
		panic("method main not found in %s\n", class_name);

	/* creation des arguments pour l'application Java */
	if( av ){
		Object *argv;
		char **pt = av;
		int i, count = 0;

		while( *pt )
			count++, pt++;

		argv = newObjectArray(string_class, count);

		for(i=0 ; i < count ; i++)
			argv->vars[i] = (int)makeJavaString(av[i]);

		setlocal(main_frame, 0, (int)argv);	
	}else{
		Object *argv;

		argv = newObjectArray(string_class, 0);

		setlocal(main_frame, 0, (int)argv);
	}


DBG1(chat("\nBegin exécution :\n");)

	for(;;){
		schedule();
		interpreter();
		if( current_frame )
			if( current_frame->exception ) /* exception non attrapée */
				dumpStackTrace(current_frame->exception);
		killThread(getCurrentThread());
	}

	panic("Je ne doit jamais passer ici\n");

DBG1(chat("\nFin d'exécution.\n");)
}


static void
dumpStackTrace(Object *thr)
{
	Object *backtrace = (Object *)thr->vars[0];
	char buf[2048], name[256];
	int i, u, count;

	sprint(buf, "%s\n", thr->class->r_name);
	slash2point(buf);
	write(2, buf, strlen(buf));

	/* construire la chaine a afficher */
	for(i = count = 0 ; i < backtrace->vars_size/sizeof(int) ; i += 2){

		for(u=0 ; (name[u] = *(((MethodBlock *)backtrace->vars[i])->name + u)) != '(' ; u++);
		name[u] = 0;
		count += sprint(buf + count, "\tat %s.%s(%s:%d)\n", 
							((MethodBlock *)backtrace->vars[i])->class->r_name,
							name,
							((MethodBlock *)backtrace->vars[i])->class->r_file,
							backtrace->vars[i+1]);
	}

	write(2, buf, strlen(buf));
}
