/* 
 * Lecture des fichiers zippés
 * J'ai réécrit ces fonctions depuis 
 * les sources du JDK de Sun.
 */


#include <u.h>
#include <libc.h>

#include "java_const.h"
#include "runtime.h"
#include "fns.h"
#include "mzip.h"

static int
mcmp(void *a, void *b)
{
	return strcmp(((zipdir *)a)->filename, ((zipdir *)b)->filename);
}

static int
findEnd(zip *z)
{
	char buf[64 + SIGSIZE], *p;
	long file_length, mark, off;

	if( (file_length = seek(z->fd, 0, 2)) < 0 )
		return 0;
	
	mark = max(0, file_length - 0xffff);
	memset(buf, 0, SIGSIZE);

	for(off = file_length ; off > mark ; ){
		int n = min(off - mark, 64);
		memcpy(buf + n, buf, SIGSIZE);
		off -= n;
		if( seek(z->fd, off, 0) < 0 )
			return 0;
		
		if( read(z->fd, buf, n) != n )
			return 0;
		
		for( p = buf + n -1 ; p >= buf ; p -= 1){
			if( !strncmp(p, ENDSIG, SIGSIZE) ){
				char endbuf[ENDHDRSIZE];
				long endoff = off + (p - buf);

				if( file_length - endoff < ENDHDRSIZE )
					continue;
				
				if( seek(z->fd, endoff, 0) < 0 )
					return 0;
				if( read(z->fd, endbuf, ENDHDRSIZE) != ENDHDRSIZE )
					return 0;
				if( endoff + ENDHDRSIZE + ENDCOM(endbuf) != file_length )
					continue;
				if( seek(z->fd, endoff, 0) < 0 )
					return 0;
				z->endoff = endoff;
				return 1;
			}
		}
	}
	return 0;
}

static int
initZip(zip *z)
{
	char buf[ENDHDRSIZE], *cenbuf, *p;
	int i, n, cenlen;

	if( !findEnd(z) )
		return 0;
	
	if( read(z->fd, buf, ENDHDRSIZE) != ENDHDRSIZE )
		return 0;

	z->cenoff = ENDOFF(buf);
	cenlen = ENDSIZ(buf);
	if( z->cenoff + cenlen != z->endoff )
		return 0;
	
	z->ndir = ENDTOT(buf);
	if( z->ndir * CENHDRSIZE > cenlen )
		return 0;

	if( ENDSUB(buf) != z->ndir )
		return 0;
	
	if( seek(z->fd, z->cenoff, 0) < 0 )
		return 0;

	/* look up table */
	z->dir = (zipdir *)malloc(z->ndir * sizeof(zipdir));
	if( !z->dir )
		return 0;
	
	/* central directory */
	cenbuf = malloc(cenlen);
	if( !cenbuf )
		return 0;
	if( read(z->fd, cenbuf, cenlen) != cenlen )
		return 0;
	
	/* Scan central directory */
	for(i=0, p = cenbuf; i < z->ndir ; i++){
		zipdir *dz = &z->dir[i];
		
		n = CENNAM(p);
		dz->filename = (char *)malloc(n + 1);
		if( !dz->filename )
			panic("malloc failed in initZip()\n");
		strncpy(dz->filename, p + CENHDRSIZE, n);
		dz->filename[n] = 0;
	
		dz->off = CENOFF(p);
		dz->len = CENLEN(p);
		
		p += CENHDRSIZE + CENNAM(p) + CENEXT(p) + CENCOM(p);
		if( p > cenbuf + cenlen )
			return 0;
	}

	free(cenbuf);
	qsort(z->dir, z->ndir, sizeof(zipdir), mcmp);
	return 1;
}

zipdir *
zipLookup(zip *zip, const char *fn)
{
	zipdir key;
	int f, c, d, r;
	key.filename = (char *)fn;

	/* dichotomie */
	d = 0;
	f = zip->ndir - 1;
	while( (f-d) > 1 ){
		c = d + ((f - d) / 2);
		r = mcmp(&key, zip->dir + c);
		if( r < 0 ) 		f = c;
		else if( r > 0 ) 	d = c;
		else 				return zip->dir + c;
	}
	return 0;
}

zip *
zipOpen(char *zip_file)
{
	int fd;
	zip *z;

	fd = open(zip_file, OREAD);
	if( fd < 0 ){
		fprint(2, "can't read file %s\n", zip_file);
		return 0;
	}
	
	z = (zip *) sysMalloc(sizeof(zip));
	if( !z )
		return 0;
	
	z->filename = strdup(zip_file);
	z->fd = fd;
	if( !initZip(z) ){
		fprint(2, "can't read zip file structure\n");
		free(z->filename);
		close(z->fd);
		free(z);
		z = 0;
	}
	close(z->fd);
	return z;
}