/*
 * Gestion des objets
 */

#include <u.h>
#include <libc.h>
#include <lock.h>

#include "java_const.h"
#include "runtime.h"
#include "fns.h"

static void resolveVars(ClassFile *, Object *);

Object *
newObject(ClassFile *class) /* arrivé ici toutes les supers sont chargées et init. */
{	
	Object *new_object;

DBG1(chat("New object Allocation (%s): ", class->r_name);)
	new_object = (Object *) sysMalloc(sizeof(Object));
	if( !new_object )
		panic("can't alloc new object in new object\n");

	new_object->class = class;
	resolveVars(class, new_object);
	new_object->type = 0;
	memset(&new_object->s, 0, sizeof(Sync));

	return new_object;
}

Object *
newString(ClassFile *string_class) 
{	
	Object *new_string;

DBG1(chat("New String Allocation (%s): ", string_class->r_name);)
	new_string = (Object *) sysMalloc(sizeof(Object));
	if( !new_string )
		panic("can't alloc new object in newString\n");

	new_string->class = string_class;
	resolveVars(string_class, new_string);
	new_string->type = T_STRING;
	memset(&new_string->s, 0, sizeof(Sync));

	return new_string;
}

Object *
newArray(int type, int n) 
{	
	Object *new_array;

DBG1(chat("New Array Allocation : ");)
	new_array = (Object *) sysMalloc(sizeof(Object));
	if( !new_array )
		panic("can't alloc new object in new object\n");

	new_array->class = 0;
	new_array->type  = type;
	switch(type){
		case T_BYTE:
			new_array->vars_size = n;
			break;
		case T_CHAR:
			new_array->vars_size = n * sizeof(short);
			break;
		case T_SHORT:
		case T_BOOLEAN:	
		case T_FLOAT:
		case T_INT:
			new_array->vars_size = n * sizeof(int);
			break;
		case T_DOUBLE:
		case T_LONG:
			new_array->vars_size = n * sizeof(vlong);
			break;
		default:
			panic("type unknown in newArray() %d\n", type);
	}
	if( new_array->vars_size < 0 )
		panic("array size < 0 in newArray()\n");
	new_array->vars = (int *)sysMalloc(new_array->vars_size);
	if( !new_array->vars )
		panic("malloc failed in newArray()\n");
	memset(new_array->vars, 0, new_array->vars_size);
	memset(&new_array->s, 0, sizeof(Sync));

DBG1(chat("%d elements %d bytes allocated\n", 
			n, new_array->vars_size); )
	return new_array;
}

Object *
newObjectArray(ClassFile *class, int n) 
{	
	Object *new_array;

DBG1(chat("New Object Array Allocation (%s): ", 
					(class != (ClassFile *)1 ) ? class->r_name : "primitive type"); )
	if( n < 0 )
		panic("object array size < 0 in newObjectArray() %d\n", n);
	new_array = (Object *) sysMalloc(sizeof(Object));
	if( !new_array )
		panic("can't alloc new object in newObjectArray()\n");

	new_array->class = class;
	new_array->type  = 1;
	new_array->vars_size = n * sizeof(int);
	new_array->vars = (int *)sysMalloc(new_array->vars_size);
	if( !new_array->vars )
		panic("malloc failed in newObjectArray()\n");
	memset(new_array->vars, 0, new_array->vars_size);
DBG1(chat("%d elements %d bytes allocated\n", 
			n, new_array->vars_size);)
	memset(&new_array->s, 0, sizeof(Sync));

	return new_array;
}

static Object*
recurMultiArray(int type, int cur_dim, int dim, int *vdim, ClassFile *class)
{
	Object *o;
	int i;

	if( cur_dim >= dim - 1 )
		if( type != T_CLASS )
			return newArray(type, vdim[cur_dim]);
		else
			return newObjectArray(class, vdim[cur_dim]);

	o = newObjectArray((ClassFile *)1, vdim[cur_dim]);
	for(i=0 ; i < vdim[cur_dim] ; i++)
		o->vars[i] = (int)recurMultiArray(type, cur_dim + 1, dim, vdim, class);
	return o;
}

Object *
newMultiArray(ClassFile *class, int index, int dimension, int *vdim)
{
	Object *new_array;	
	ClassFile *target_class;
	cp_info *cp = &class->constant_pool[index];
	char sign[256], *pt;
	int type;
DBG1(int i;)
	
DBG1(chat("New Multi Object Array Allocation : dim = %d (", dimension);
	for(i=0 ; i < dimension ; i++ )
		chat("%d ", vdim[i]);
	chat(")\n"); )
	GetCpString(sign, cp->u.c.name_index, class);
	pt = sign;
	while( *pt == '[' ) pt++;

	type = sign2type(pt);
	if( type == T_CLASS ){
		pt++; pt[strlen(pt)-1] = '\0';
		target_class = lookupClass(pt);
	}else
		target_class = 0;

	new_array = recurMultiArray(type, 0, dimension, vdim, target_class);
	memset(&new_array->s, 0, sizeof(Sync));

	return new_array;
}

int
arrayLength(Object *array)
{

	if( IS_ARRAY(array) ) /* tableau (int, long ...) */	
		switch( array->type ){
			case T_BYTE:
				return array->vars_size;
			case T_CHAR:
				return array->vars_size / sizeof(short);
			case T_SHORT:
			case T_BOOLEAN:
			case T_FLOAT:
			case T_INT:
				return array->vars_size / sizeof(int);
			case T_DOUBLE:
			case T_LONG:
				return array->vars_size / sizeof(vlong);
			default:
				panic("type unknown in arrayLength() %d\n", array->type);	
		}
	else if( IS_ARRAY_OBJECT(array) ) /* tableau d'objets */
		return array->vars_size / sizeof(int);
	else
		panic("not an array arrayLength()\n");

	return -1;
}


static void
resolveVars(ClassFile *class, Object *object)
{
	ClassFile *cur_class;
	cp_info *cp;
	field_info *fi;
	char signature[256];
DBG1(char buf[256];)
	int i, nvars;

	if( !class )
		return;

	cur_class = class;
	nvars = 0;
	while( cur_class ){
		/* calcul la taile du nombre de variables d'instance */
		for(i=0 ; i < cur_class->fields_count ; i++){
			fi = &cur_class->fields[i];
			if( (fi->access_flag&ACC_STATIC) )
				continue;
			cp = &cur_class->constant_pool[fi->signature_index];
			strncpy(signature, (char *)cp->u.a.bytes, cp->u.a.length);
			signature[cp->u.a.length] = 0;
DBG1(		GetCpString(buf, fi->name_index, cur_class );
			chat("<%s (%s)>", buf, signature); )
			nvars += signatureSize(signature);
		}
		cur_class = cur_class->r_super;
	}
DBG1(chat("nvars = %d\n", nvars);)
	object->vars = (int *) sysMalloc(nvars*sizeof(int));
	if( !object->vars )
		panic("sysMalloc failed in resolveVars()\n");
	memset(object->vars, 0, nvars*sizeof(int));
	object->vars_size = nvars;
}

void 
javaString2CString(Object *obj_string, char *buf, int buflen)
{
	Object *tab_char;
	short *pt;
	int i, count, offset;

	/*if( obj_string->type != T_STRING )
		panic("not a object string in javaString2CString()\n");*/

	tab_char = (Object *)obj_string->vars[0];
	offset	 = obj_string->vars[1];
	count	 = obj_string->vars[2];
	
	if( count >= buflen )
		panic("buf too small in javaString2CString()\n");

	pt = (short *)tab_char->vars;

	for(i=0; i < count ; i++){
		buf[i] = pt[i+offset];
	}
	buf[count] = 0;
}

Object *
makeJavaString(char *str)
{
	ClassFile *c;
	Object *string, *tab_char;
	int i, len = strlen(str);
	
	/* classe String */
	c = getClass("java/lang/String");

	/* création de l'objet et resolution */
	string = newString(c);
			
	/* creation du tableau de char contenant la chaine */
	tab_char = newArray(T_CHAR, len);
	for(i=0 ; i < len ; i++)
		((short *)tab_char->vars)[i] = 	str[i];

	executeConstructor(string, "([C)V", tab_char);
	return string;
}

void 
setProperty(Object *prop, char *type, char *value)
{
	ClassFile *hash_class = getClass("java/util/Hashtable");

	if( !hash_class )
		panic("couldn't get class HashTable !!!! \n");

	/* On lance ici la méthode Hashtable.put à
	 * travers l'objet Properties (invokevirtual)
	 */

	executeVirtual(hash_class, prop, "put",
					"(Ljava/lang/Object;Ljava/lang/Object;)Ljava/lang/Object;",
					makeJavaString(type), makeJavaString(value));
}


int
sign2type(char *sign)
{
	switch(*sign){
		case 'I':
			return T_INT;
		case 'Z':
			return T_BOOLEAN;
		case 'J':
			return T_LONG;
		case 'V':
			return T_VOID;
		case 'F':
			return T_FLOAT;
		case 'D':
			return T_DOUBLE;
		case 'L':
			return T_CLASS;
		case '[':
			return T_TAB;
		case 'B':
			return T_BYTE;
		case 'C':
			return T_CHAR;
	default:
		panic("unknown type in sign2type() %s\n", sign);
	}

	return -1;
}

int 
typeSize(int type){
	
	switch(type){
		case T_CLASS: 	/* classe  */
		case T_TAB: 	/* tableau */
		case T_INT: 	/* integer */
		case T_FLOAT: 	/* float   */
		case T_BOOLEAN: /* boolean */
		case T_CHAR: 	/* char    */
			return 1;
		case T_DOUBLE: 	/* double  */
		case T_LONG: 	/* long    */
			return 2;
		default:
				panic("unknown type %d typeSize()\n", type);
	}

	return -1;
}