/*
 * Constant Pool Resolution et toute la clique
 */

#include <u.h>
#include <libc.h>
#include <lock.h>

#include "java_const.h"
#include "runtime.h"
#include "fns.h"


int syn_lock = 0;

static void preparation(ClassFile *class);

ClassFile *
lookupClass(char *class_name)
{
	ClassFile *c;
	
	//syn_lock++;
	//getLock(&lookup_sync);

	c = loadClass(class_name); 
	if( !c ){
		return 0;
	}
	linking(c); 
	initialize(c);

	//syn_lock--;
	//releaseLock(&lookup_sync);

	return c;
}


/* 
 * Linked => verified, prepared and resolved
 * here only preparation is done.
 */
void
linking(ClassFile *cur_class)
{
	if( cur_class->r_state&CF_LINKED )
		return;

	if( cur_class->r_super )
		linking(cur_class->r_super);

	preparation(cur_class);

	cur_class->r_state |= CF_LINKED;
}

static Sync lookup_sync;


static void
preparation(ClassFile *class)
{
	field_info *fi;
	Object *object;
	char signature[256], name[256], buf[256];
	cp_info *cp;
	int i, j, n, u, n_methods, flag;
	
	/* Fields */

DBG1(chat("Preparation for %s:\n\t", class->r_name);)
	
	for(n = 0, i=0 ; i < class->fields_count ; i++){
		fi = &class->fields[i];
		if( !(fi->access_flag&ACC_STATIC) )
			continue;
		GetCpString(signature, fi->signature_index, class);
DBG1(	GetCpString(buf, fi->name_index, class);
		chat("<%s (%s)>", buf, signature); )
		n += signatureSize(signature);
	}

	class->r_fields = 0;
	if( n ){
		class->r_fields = (int *)sysMalloc(n*sizeof(int));
		if( !class->r_fields )
			panic("sysMalloc failed in preparation()\n");
DBG1(	chat("%d bytes allocated at %d\n", n*sizeof(int), class->r_fields); )
	
		/* resolution and init for static fields */

		memset(class->r_fields, 0, n*sizeof(int));
		for(n = 0, i=0 ; i < class->fields_count ; i++){
			fi = &class->fields[i];
DBG1(		GetCpString(buf, fi->name_index, class);
			chat("<%s>\n", buf); )
			if( !(fi->access_flag&ACC_STATIC) )
				continue;
			for(j=1 ; j < class->constant_pool_count ; j++){
				cp = &class->constant_pool[j];
				if( cp->tag == CONSTANT_Fieldref ){
					cp_info *cp_nt = &class->constant_pool[cp->u.m.name_and_type_index];
					if( (cp_nt->u.n.name_index == fi->name_index) && 
						(cp_nt->u.n.signature_index == fi->signature_index) &&
						(cp->u.m.class_index == 1) /* resolution pour la classe uniquement */){
						GetCpString(signature, fi->signature_index, class);
						break;
					}
				}
			}
			/*if( j >= class->constant_pool_count ){
				chat("can't find constant pool entry in preparation\n");
				n += signatureSize(signature);
				continue;
			}*/
			if( j < class->constant_pool_count ){
DBG1(			chat("resolution %d...\n", j); )
				class->r_resolved[j] = (int)(class->r_fields + n);
			}
			n += signatureSize(signature);
		}
	}

	/* Methods */

	//fprint(2, "class = %s\n", class->r_name);
	if( class->r_super ){
		class->r_methods = (MethodBlock *)sysMalloc(class->r_super->r_n_methods *
													sizeof(MethodBlock));
		memcpy(class->r_methods, class->r_super->r_methods,
				class->r_super->r_n_methods * sizeof(MethodBlock));
		n_methods = class->r_super->r_n_methods;
	}else{
		class->r_methods = 0;
		n_methods = 0;
	}
	//print("n_methods = %d\n", n_methods);
	for(n=0 ; n < class->methods_count ; n++){
		GetCpString(signature, class->methods[n].signature_index, class);
		GetCpString(name, class->methods[n].name_index, class);
		sprint(buf, "%s%s", name, signature);	
		//print("--> %s ", buf);

		flag = 0;
		if( class->r_super ){
			for(i=0 ; i < class->r_super->r_n_methods ; i++){
				//print("%s == %s\n", buf, class->r_methods[i].name);
				if( !strcmp(buf, class->r_methods[i].name) && 
						strncmp(buf, "<init>", 6) && 
						strncmp(buf, "<clinit>", 8) ){
					//print("overide %d\n", i);
					class->r_methods[i].class = class;
					class->r_methods[i].access_flag = class->methods[n].access_flag;
					for (u=0; u < class->methods[n].attributes_count ; u++)
						if (class->methods[n].attributes[u].type == ATTR_CODE) 
							class->r_methods[i].ca = class->methods[n].attributes[u].u.ca;
					break;
				}
			}
			if( i < class->r_super->r_n_methods )
				flag = 1; 
		}
		if( flag == 0 ){
			class->r_methods = (MethodBlock *)realloc(class->r_methods,
													(n_methods + 1) * sizeof(MethodBlock));
			class->r_methods[n_methods].class = class;
			class->r_methods[n_methods].access_flag = class->methods[n].access_flag;
			class->r_methods[n_methods].nargs = argSize(signature);	
			class->r_methods[n_methods].name = strdup(buf);
			for (i=0; i < class->methods[n].attributes_count ; i++)
				if (class->methods[n].attributes[i].type == ATTR_CODE) 
					class->r_methods[n_methods].ca = class->methods[n].attributes[i].u.ca;
			//print("added at %d (%d)\n", n_methods, flag);
			n_methods++;
		}
	}
	class->r_n_methods = n_methods;
	
/*	for(n=0 ; n < n_methods ; n++){
		MethodBlock *p = class->r_methods + n;
		fprint(2, "%s -> %s\n", p->name, p->class->r_name);
	}
	fprint(2, "\n");
*/
	/* Creation de la classe Class */
	
	object = newObject(getClass("java/lang/Class"));
	object->vars = (int *)sysMalloc(sizeof(int));
	object->vars_size = 1;
	object->vars[0] = (int)class;
	class->r_java_class = object;
}

void
initialize(ClassFile *cur_class)
{
	Frame *frame;

	if( cur_class->r_state&(CF_INITIALIZED|CF_BEING_INITIALIZED))
		return;
	
	cur_class->r_state |= CF_BEING_INITIALIZED;

	if( cur_class->r_super )
		initialize(cur_class->r_super);
	
DBG1(chat("Initialization of %s :\n", cur_class->r_name);)
	frame = newFrameFromName(cur_class, "<clinit>", "()V");
	if( frame ){
		interpreter();
		freeFrame(frame);
DBG1(	chat("\t<clinit> method in %s class executed\n\n", cur_class->r_name); )
	}
DBG1(else
		chat("\tNo clinit method in %s class\n\n", cur_class->r_name);)

	cur_class->r_state |= CF_INITIALIZED;
	cur_class->r_state &= ~CF_BEING_INITIALIZED;
}

static Sync resolution_sync;

void
cp_resolution(ClassFile *class, int index, Object *object)
{
	char buf[512];
	cp_info *cp = &class->constant_pool[index];
	
	//syn_lock++;
	//getLock(&resolution_sync);

	switch(cp->tag){
		case CONSTANT_Class: { /* ClassFile * in r_resolved */
			ClassFile *c; 

			GetCpString(buf, cp->u.c.name_index, class);
			if( *buf == '[' ){ /* un tablo de classe ou de type primitif*/
				class->r_resolved[index] = -1;	
			}else{ 
				c = lookupClass(buf); 
				if( !c ){
					throwException(getCurrentFrame(), "java/lang/NoClassDefFoundError");
					break;
				}
				//print("%s %d %d\n", buf, index, c);
				class->r_resolved[index] = (int)c;
			}
			
			break;
		}
		case CONSTANT_Fieldref: {
			char sign[256], name[256], o_sign[256], o_name[256];
			ClassFile *cur_class;
			cp_info *cp1;
			int n, nvars;
			
			if( !object ){ 	/* variables de classe (getstatic et putstatic) */
				/* recherche de la classe contenant la var. statique */
				if( !class->r_resolved[cp->u.m.class_index]	)
					cp_resolution(class, cp->u.m.class_index, 0);
				cur_class = (ClassFile *)class->r_resolved[cp->u.m.class_index];

				/* recupere le nom du champ et sa signature */
				cp1 = &class->constant_pool[cp->u.m.name_and_type_index];
				GetCpString(name, cp1->u.n.name_index, class);
				GetCpString(sign, cp1->u.n.signature_index, class);

				/* ou est la variable dans la classe */

				for (nvars = 0, n=0; n < cur_class->fields_count ; n++){
					if( !(cur_class->fields[n].access_flag&ACC_STATIC) )
						continue;
					GetCpString(o_name, cur_class->fields[n].name_index, cur_class);
					GetCpString(o_sign, cur_class->fields[n].signature_index, cur_class);
					if( !strcmp(name, o_name) && !strcmp(sign, o_sign) ){
						/* le bon champ est trouvé on le resoud */
						class->r_resolved[index] = (int)(cur_class->r_fields + nvars);
						//releaseLock(&resolution_sync);
						//syn_lock--;
						return;
					}
					nvars += signatureSize(o_sign);
				}
				throwException(getCurrentFrame(), "java/lang/NoSuchFieldError");
			}else{			/* variables d'instances (getfield et putfield) */
			
				/* recupere le non du champ et sa signature */
				cp1 = &class->constant_pool[cp->u.m.name_and_type_index];
				GetCpString(name, cp1->u.n.name_index, class);
				GetCpString(sign, cp1->u.n.signature_index, class);

				/* recupere la classe du champ */
				cur_class = object->class;

				nvars = 0;
				while( cur_class ){
					/*  trouve le bon champ */
					for (n=0; n < cur_class->fields_count ; n++){
						if( cur_class->fields[n].access_flag&ACC_STATIC )
							continue;
						GetCpString(o_name, cur_class->fields[n].name_index, cur_class);
						GetCpString(o_sign, cur_class->fields[n].signature_index, cur_class);
						if( !strcmp(name, o_name) && !strcmp(sign, o_sign) ){
							/* le bon champ est trouvé on le resoud */
							class->r_resolved[index] = nvars + 1;
							//releaseLock(&resolution_sync);
							//syn_lock--;
							return;
						}
						nvars += signatureSize(o_sign);
					}
				
					cur_class = cur_class->r_super;
				}
				throwException(getCurrentFrame(), "java/lang/NoSuchFieldError");
			}
			break;
		}
		case CONSTANT_Methodref: { /* MethodBlock or offset in r_resolved */
			char sign[256], name[256];
			ClassFile *c;
			cp_info *cp1;
			int n;

			/* recupere le non de la methode et sa signature */
			cp1 = &class->constant_pool[cp->u.m.name_and_type_index];
			GetCpString(name, cp1->u.n.name_index, class);
			GetCpString(sign, cp1->u.n.signature_index, class);
			sprint(buf, "%s%s", name, sign);
			//print("cp_resolution : %s\n", buf);

			/* recupere la classe de la methode */
			if( !class->r_resolved[cp->u.m.class_index] ){
				//print("class index 1 = %d\n", cp->u.m.class_index);
				cp_resolution(class, cp->u.m.class_index, 0);
			}
			//print("class index 2 = %d\n", cp->u.m.class_index);
			c = (ClassFile *)class->r_resolved[cp->u.m.class_index];
			if( !c ){ 
				panic("class not found %d\n", cp->u.m.class_index);
			}

			for (n=c->r_n_methods-1; n >= 0  ; n--){
				MethodBlock *p = c->r_methods + n;
				if( !strcmp(buf, p->name) ){
					if( object )
						class->r_resolved[index] = ((n<<16)|(p->nargs&0xFFFF));
					else
						class->r_resolved[index] = (int)p;
					//syn_lock--;
					return;
				}
			}
			panic("can't resolved methodref %s in cp_resolution()\n", buf);
			break;
		}
		case CONSTANT_String: {
			ClassFile *c;
			Object *string, *tab_char;
			cp_info *cp2;
			int i;

			/* classe String */
			c = lookupClass("java/lang/String"); 

			/* création de l'objet et resolution */
			string = newString(c);
			class->r_resolved[index] = (int)string;
			
			/* creation du tableau de char contenant la chaine */
			cp2 = &class->constant_pool[cp->u.s.string_index];
			tab_char = newArray(T_CHAR, cp2->u.a.length);
			//memcpy((char *)tab_char->vars, (char *)cp2->u.a.bytes, cp2->u.a.length);
			for(i=0 ; i < cp2->u.a.length ; i++)
				((short *)tab_char->vars)[i] = 	cp2->u.a.bytes[i];

			/* exécution du constructeur */
			executeConstructor(string, "([C)V", tab_char);

			break;
		}
		default:
			panic("resolution type not yet imlpemented\n");
	}
	//releaseLock(&resolution_sync);
	//syn_lock--;
}


int
isSubClassOf(ClassFile *src, ClassFile *dst)
{
	while( src ){
		if( src == dst )
			return 1;
		src = src->r_super;
	}
	return 0;
}

int
isImplementOf(ClassFile *src, ClassFile *dst)
{
	ClassFile *c;
	int i;

	for(i=0 ; i < src->interfaces_count ; i++){
		if( !src->r_resolved[src->interfaces[i]] )
			cp_resolution(src, src->interfaces[i], 0);
		c = (ClassFile *)src->r_resolved[src->interfaces[i]];
			
		if( c == dst )
			return 1;
	}

	return 0;	
}

int
isClassArray(int index, ClassFile *class)
{
	char buf[256];
	cp_info *cp = &class->constant_pool[index];

DBG1(if( cp->tag != CONSTANT_Class )
		panic("bad type in isClassArray()\n"); )

	GetCpString(buf, cp->u.c.name_index, class);

	if( *buf == '[' )
		return 1;

	return 0;
}

int
classArrayType(int index, ClassFile *class)
{
	char buf[256], *pt;
	cp_info *cp = &class->constant_pool[index];

DBG1(if( cp->tag != CONSTANT_Class )
		panic("bad type in isClassArray()\n"); )

	GetCpString(buf, cp->u.c.name_index, class);
	
	pt = buf;
	while( *pt == '[' ) pt++;
	
	return sign2type(pt);
}

void 
getSuperClassName(char *s, ClassFile *class)
{
	cp_info *c,*cp; 

	if( !class->super_class ){
		s[0] = 0;
		return;
	}

	c = &class->constant_pool[class->super_class];
	cp = &class->constant_pool[c->u.c.name_index];

	strncpy(s, (char *)cp->u.a.bytes, cp->u.a.length);
	s[cp->u.a.length] = 0;
}


int 
signatureSize(char *signature)
{

	switch( *signature ){
		case 'L': /* classe  */
		case '[': /* tableau */
		case 'I': /* integer */
		case 'F': /* float   */
		case 'Z': /* boolean */
		case 'C': /* char    */
		case 'B': /* Byte    */
		case 'S': /* short   */
			return 1;
		case 'D': /* double  */
		case 'J': /* long    */
			return 2;
		default:
				panic("unknown signature %s signatureSize()\n", signature);
	}

	return 0;
}

int
argSize(char *signature)
{
	int nargs;
	char *p = signature;	

	while( *p++ != '(' );

	for(nargs = 0; *p != ')' && *p ; p++)
		switch( *p ){
			case '[': /* tableau */
				nargs += 1;
				while( *p == '[' ) p++;
				if( *p == 'L' ) 
					while( *p != ';' ) p++;
				break;
			case 'L': /* classe */
				nargs += 1;
				while( *p != ';' ) p++;
				break;
			case 'I': /* integer */
			case 'F': /* float   */
			case 'Z': /* boolean */
			case 'C': /* char    */
			case 'S': /* short   */
			case 'B': /* byte	 */
				nargs += 1;
				break;
			case 'D': /* double */
			case 'J': /* long */
				nargs += 2;
				break;
			default:
				panic("unknown signature %s argSize()\n", signature);
		}
	
	return nargs;
}

/*void
GetCpString(char *string, int n, ClassFile *class)
{
	int i;
	uchar *bytes = class->constant_pool[n].u.a.bytes;

	/*if( class->constant_pool[n].tag != CONSTANT_Asciz )
		panic("GetCpString() not CONSTANT_Asciz (%d) for entry %d\n", 
			   class->constant_pool[n].tag, n);
	
	for (i=0 ; i < class->constant_pool[n].u.a.length; i++)
		string[i] = *(bytes + i);
	string[i] = 0;
}*/

void
getClassNameFromCP(char *buf, int index, ClassFile *class) /* method, field */
{
	cp_info *cp = &class->constant_pool[index];
	cp_info *cp1;


DBG1(if( !((cp->tag == CONSTANT_Fieldref) || 
			(cp->tag == CONSTANT_Methodref) ||
				(cp->tag == CONSTANT_InterfaceMethodref)) )
		panic("bad type in getClassNameFromCP() %d\n", cp->tag); )

	cp1 = &class->constant_pool[cp->u.m.class_index];

	GetCpString(buf, cp1->u.c.name_index, class);
}

void
getNameFromCP(char *buf, int index, ClassFile *class) /* method, field */
{
	cp_info *cp = &class->constant_pool[index];
	cp_info *cp1;

DBG1(if( !((cp->tag == CONSTANT_Fieldref) ||
			(cp->tag == CONSTANT_Methodref) || 
				(cp->tag == CONSTANT_InterfaceMethodref)) )
		panic("bad type in getMethodNameFromCP() %d\n", cp->tag); )

	cp1 = &class->constant_pool[cp->u.m.name_and_type_index];
	
	GetCpString(buf, cp1->u.n.name_index, class);
}

void
getSignatureFromCP(char *buf, int index, ClassFile *class) /* method, field */
{
	cp_info *cp = &class->constant_pool[index];
	cp_info *cp1;

DBG1(if( !((cp->tag == CONSTANT_Fieldref) ||
			(cp->tag == CONSTANT_Methodref) ||
				(cp->tag == CONSTANT_InterfaceMethodref)) )
		panic("bad type in getSignatureFromCP()\n"); )

	cp1 = &class->constant_pool[cp->u.m.name_and_type_index];
	
	GetCpString(buf, cp1->u.n.signature_index, class);
}

void
getClassFromCP(char *buf, int index, ClassFile *class) /* CONSTANT_Class */
{
	cp_info *cp = &class->constant_pool[index];

DBG1(if( cp->tag != CONSTANT_Class )
		panic("bad type in getClassFromCP()\n"); )

	GetCpString(buf, cp->u.c.name_index, class);
}

int 
isCompatibleObject(ClassFile *type, ClassFile *required)
{
	if( type->access_flags&ACC_INTERFACE ){
		//print("interface\n");
		return isImplementOf(type, required);
	}
	
	return isSubClassOf(type, required);
}

int 
isCompatibleArray(Object *obj, char *rtype)
{
	char *pt = rtype;
	ClassFile *rclass = 0;

	if( *rtype != '[' )
		panic("Hiiiiiii\n");

	while( *pt == '[' ) pt++;
	if( *pt == 'L' ){
		char *p = pt;
		while( *p != ';' ) p++;
		*p = 0;
		rclass = lookupClass(pt+1);
		if( !rclass ) panic("can't find class %s in isCompatibleArray()\n", pt+1);
	}else
		print("it's a primitive type <%s><%s> in isCompatibleArray()\n", pt,rtype);
	
	if( rclass->access_flags&ACC_INTERFACE ){
		panic("interface in isCompatibleArray()\n");
	}else if( rclass ){ /* c'est une classe */
		if( obj->class == rclass || obj->class == getClass("java.lang.Object") ||
			isSubClassOf(obj->class, rclass) )
			return 1;
		print("une classe\n");
	}else{ /*c'est un tableau de types primitifs */
		print("buf = %s\n", rtype);
		panic("primitifs in isCompatibleArray()\n");
		
	}

	return 0;
}
