/*
 * Declaration runtime
 */

/* 
 * trace les appels de fonctions 
 */
//#define DBG_1
/* 
 * DBG_1 + les instructions 
 */
//#define DBG_2
/* 
 * DBG_2 + le contenue des frames 
 */
//#define DBG_3
/* 
 * histograme des instructions java 
 */
//#define PROFILE

/*
 * LITTLE_ENDIAN for x86, Pentiums... 
 * undef for big endian bytes order sparc, 68000 ...
 */

#define LITTLE_ENDIAN

typedef struct Frame Frame;
typedef struct StackFrame StackFrame;
typedef struct MethodBlock MethodBlock;
typedef struct Object Object;
typedef struct thContext thContext;
typedef struct Sync Sync;
typedef struct zip zip;

#define MB_MAX_LOCALS(m)		((m)->ca->max_locals)
#define MB_MAX_STACK(m)			((m)->ca->max_stack)
#define MB_CODE(m)				((m)->ca->code)

struct MethodBlock {
	ClassFile *class;	/* classe de la méthode */
	char *name;			/* nom + signature */
	int nargs;			/* nombre d'arguments */
	int access_flag;	/* bits d'access */
	Code_attribute *ca; /* attributs de code */
};


#define FRAME_CA(f)				((f)->mb->ca)
#define FRAME_MAX_STACK(f)		((int)MB_MAX_STACK((f)->mb))
#define FRAME_MAX_LOCALS(f)		((int)MB_MAX_LOCALS((f)->mb))

struct Frame {
	uchar *pc;			/* compteur de programme */
	int *locals;		/* registres pour les locales */
	int *opstack;		/* pile d'opérande */
	int opstack_pt;		/* pointeur de pile */
	Object *exception;	/* exception lancée ? */
	Object *obj_lock;	/* si il y a synchro sur un objet */

	ClassFile *class;	/* classe de la méthode	*/
	Frame *invoker;		/* frame de la méthode appelante */
	char *name;			/* nom + signature */
	MethodBlock *mb;	/* pt. sur la descripteur de méthode */
};


struct StackFrame {
	Frame **stack;
	int pt;
	int max;
	int size;
};

struct Sync {
	int ref;
	Object *thread_owner;
	thContext *wait;
	thContext *event;
};

struct Object {
	ClassFile *class;	/* classes de l'objet */
	int type;			/* type de l'objet */
	int *vars;			/* variables d'instance ou espace pour les tab. */
	int vars_size;		/* taille du champs vars */
	Sync s;
};

#define IS_OBJECT(a)			((a)->class && (!(a)->type || ((a)->type == T_STRING)))
#define IS_ARRAY(a)				(!(a)->class && (a)->type)
#define IS_ARRAY_OBJECT(a)		((a)->class && ((a)->type == T_TAB))

/* Definitions pour les listes d'arguments variables */

typedef char *arg_list;
#define arg_begin(list, last)		((list) = (char *)(&(last) + 1))
#define arg_value(list, mode)		((mode *)(list += sizeof(mode)))[-1]
#define arg_end(list)				


#define WORD	1
#define DWORD	2*WORD

/* fonctions inline !!! */

#define push(f, v)			(f)->opstack[++(f)->opstack_pt] = (v)
#define pushFloat(f, v)		*((float *)(f)->opstack + ++(f)->opstack_pt) = (v)

#define pushLong(f, v)		{ 												\
								union { vlong l; uint t[2]; } u; 			\
								u.l = (v);									\
								(f)->opstack[++(f)->opstack_pt] = u.t[0];	\
								(f)->opstack[++(f)->opstack_pt] = u.t[1];	\
							}												

#define pushDouble(f, v)	{ 												\
								union { double d; uint t[2]; } u; 			\
								u.d = (v);									\
								(f)->opstack[++(f)->opstack_pt] = u.t[0];	\
								(f)->opstack[++(f)->opstack_pt] = u.t[1];	\
							}												

#define pop(f)			(f)->opstack[(f)->opstack_pt--]
#define popFloat(f)		*((float *)(f)->opstack + (f)->opstack_pt--)


#define setlocal(f, n, v)	(f)->locals[(n)] = (v)
#define getlocal(f, n)		(f)->locals[(n)]

#define setlocalFloat(f, n, v)	*((float *)((f)->locals + (n))) = (v)
#define getlocalFloat(f, n)		*((float *)((f)->locals + (n)))

#define setlocalLong(f, n, v)	*((vlong *)((f)->locals + (n))) = (v)
#define getlocalLong(f, n)		*((vlong *)((f)->locals + (n)))

#define setlocalDouble(f, n, v)	*((double *)((f)->locals + (n))) = (v)
#define getlocalDouble(f, n)	*((double *)((f)->locals + (n)))

#define GetCpString(string, n, class) { 								\
	uchar *bytes = (class)->constant_pool[(n)].u.a.bytes;				\
	int z, length = (class)->constant_pool[(n)].u.a.length;				\
	for (z=0 ; z < length; z++)											\
		*(((char *)string) + z) = *(bytes + z);							\
	*(((char *)string) + z) = 0;										\
} 																		\


/* les fonctions de debug */


#ifdef DBG_3
#define DBG3(s)		s
#else
#define DBG3(s)
#endif

#ifdef DBG_2
#define DBG2(s)		s
#else
#define DBG2(s)
#endif

#ifdef DBG_1
#define DBG1(s)		s
#else
#define DBG1(s)
#endif
