#include <u.h>
#include <libc.h>
#include <lock.h>

#include "java_const.h"
#include "runtime.h"
#include "fns.h"
#include "thread.h"

typedef struct Alarm Alarm;

struct Alarm {
	thContext *thread;
	ulong expire;
	int data;
	Alarm *next;
};

extern StackFrame *current_stack;
extern Frame *current_frame;
extern int syn_lock;

int need_schedule;
Object *current_thread;

static thContext *runQ[MAX_PRIO + 1];	/* running queue */
static thContext *suspendQ; 			/* queue for suspending thread */
static Alarm *alarmList;				/* queue for sleeping thread */

static void addQueue(thContext **, thContext *);
static void rmQueue(thContext **, thContext *);
static void addLastQueue(thContext **, thContext *);
static Thread *findThread(Object *);
static void recordAlarm(Object *, ulong, int);
static void checkAlarm(void);
void alarmHandler(void *, char *);

#define ASK_SCHEDULE()		need_schedule = 1

void
dumpAlarm(void)
{	
	Alarm *pt = alarmList;
	thContext *ctx;

	print("alarm queue : ");
	while( pt ){
		ctx = pt->thread;
		print("%d(%d), ", ctx->thread, pt->expire);
		pt = pt->next;
	}
	print("\n");
}


void
dumpQ(thContext **head)
{
	thContext *pt = *head;

	while( pt ){
		print("th(%d, %d) ", pt->thread, pt->state);
		pt = pt->next;
	}
	print("\n");
}

void dumpRunQ(void)
{
	int i;

	for(i=MAX_PRIO ; i >= MIN_PRIO ; i--){
		print("%d : ");
		dumpQ(&runQ[i]);
	}

}

void
initThread(void)
{
	Object *thread_object, *tab_char, *grp_object;
	Thread *new_thread;
	ClassFile *thread_class, *thgrp_class;
	thContext *ctx;
	int i;

	/* enrgistrement des alarmes */
	notify(alarmHandler);

	/* création du thread principal */
	thread_class = lookupClass("java/lang/Thread");
	if( !thread_class )
		panic("can't find java.lang.Thread\n");
	thread_object = newObject(thread_class);

	/* construction du contexte */
	ctx = (thContext *)sysCalloc(sizeof(thContext));
	if( !ctx )
		panic("calloc failed in initThread()\n");
	ctx->stack 		= initGlobalStack();
	ctx->state 		= TH_READY;
	ctx->prio 		= NORM_PRIO;
	ctx->thread		= thread_object;


	/* positionnement des variables de l'objet */
	new_thread = findThread(thread_object);	
	tab_char = newArray(T_CHAR, strlen("main"));
	for(i=0 ; i < strlen("main") ; i++)
		((short *)tab_char->vars)[i] = 	*((char *)"main" + i);

	new_thread->name 		= tab_char;
	new_thread->priority 	= NORM_PRIO;
	new_thread->threadQ		= 0;
	new_thread->private		= (int)ctx;
	new_thread->single_step = 0;
	new_thread->deamon 		= 0;
	new_thread->stillborn 	= 0;
	new_thread->target 		= 0;
		
	/* création du groupe */
	thgrp_class = lookupClass("java/lang/ThreadGroup");
	if( !thgrp_class )
		panic("can't find java.lang.ThreadGroup\n");
	grp_object = newObject(thgrp_class);
	current_stack = ctx->stack;
	executeConstructor(grp_object, "()V");
	executeVirtual(thgrp_class, grp_object,
					"add", "(Ljava/lang/Thread;)V", thread_object);
	
	new_thread->group 		= grp_object;

	/* place le thread pour son exécution */
	
	addQueue(&runQ[NORM_PRIO], ctx);	
}

void 
schedule(void)
{
	int i;
	thContext *ctx, *cur;
	Thread *t;
	
begin:
	need_schedule = 0;
	checkAlarm();

	for(i = MAX_PRIO ; i >= MIN_PRIO ; i--){
		if( runQ[i] ){
			ctx = runQ[i];
			
			if( current_thread ){
				t = findThread(current_thread);
				cur = (thContext *)t->private;
				if( cur->state == TH_RUNNING )
					cur->state = TH_READY;
			}

			if( ctx->state == TH_READY_LOCK ){
				Object *obj = (Object *)ctx->data;
				
				current_thread = ctx->thread;
				if( getLock(obj) < 0 )
					goto begin;
			}

			ctx->state 		= TH_RUNNING;
			current_thread 	= ctx->thread;
			current_stack 	= ctx->stack;
			current_frame 	= current_stack->stack[current_stack->pt];
			return;
		}
	}
	current_thread = 0; current_stack = 0; current_frame = 0;
	if( suspendQ || alarmList ){
		while( !need_schedule )	
			sleep(0);
		goto begin;
	}

#ifdef PROFILE
		dumpProfile();
#endif
	//abort();
	exits("ok");
}
/*
void 
schedule(void)
{
	int i;
	thContext *ctx, *cur;
	Thread *t;
	
/*	if( syn_lock )
		panic("Hiiiiiiiiiiiiiiiii = %d\n", syn_lock);

begin:
	//dumpRunQ();
	need_schedule = 0;
	checkAlarm();
	for(i = MAX_PRIO ; i >= MIN_PRIO ; i--){
		if( runQ[i] ){
			ctx = runQ[i];
			
			if( current_thread ){
				//print("here\n");
				t = findThread(current_thread);
				cur = (thContext *)t->private;
				if( cur->state == TH_RUNNING )
					cur->state = TH_READY;
			}

			if( ctx->state == TH_READY_LOCK ){
				Object *obj = (Object *)ctx->data;
				
				ctx->state 		= TH_RUNNING;
				current_thread 	= ctx->thread;
				current_stack 	= ctx->stack;
				current_frame 	= current_stack->stack[current_stack->pt];
				if( getLock(obj) < 0 )
					goto begin;
				//print("schedule (2) thread %d\n", ctx->thread);
				//dumpRunQ();
				return;
			}

			ctx->state 		= TH_RUNNING;
			current_thread 	= ctx->thread;
			current_stack 	= ctx->stack;
			current_frame 	= current_stack->stack[current_stack->pt];
			//print("schedule (1) thread %d\n", ctx->thread);
			//dumpRunQ();
			return;
		}
	}
	current_thread = 0; current_stack = 0; current_frame  = 0;
	if( suspendQ || alarmList ){
		while( !need_schedule )	
			sleep(0);
		goto begin;
	}

	print("no thread exiting\n");
	exits("ok");
}
*/
void
killThread(Object *thread)
{
	Thread *t;
	thContext *ctx;

	if( !thread )
		return;

	t = findThread(thread);
	ctx = (thContext *)t->private;

	switch(ctx->state){
		case TH_RUNNING:
		case TH_READY:
			rmQueue(&runQ[ctx->prio], ctx);
			ctx->state = TH_DEAMON;
			break;
		case TH_SLEEP:
		case TH_SUSPEND:
		default:
			panic("bad thread state in killThread()\n");
	}
}

void 
rmFromAlarmList(thContext *ctx)
{
	Alarm *pt;
	thContext *pctx;
	ulong cur_time;

	if( !alarmList )
		return;

	pt = alarmList;
	pctx = alarmList->thread;

	if( pctx == ctx ){
		alarmList = pt->next;
		free(pt);
	}else{
		
		while( pt->next ){
			if( pt->next->thread == ctx )
				break;
			pt = pt->next;
		}
		pt->next = pt->next->next;
		free(pt->next);
	}
	if( alarmList ){
		cur_time = getMiliSec();
		alarm(alarmList->expire - cur_time);
	}
}

static void
recordAlarm(Object *thread, ulong msec, int data)
{
	Thread *t; thContext *ctx;
	Alarm *pre, *pt, *a;
	ulong cur_time, expire;

	if( msec == 0 ){
		return;
	}

	cur_time = getMiliSec();
	expire = cur_time + msec;

	t = findThread(thread);
	ctx = (thContext *)t->private;

	a = (Alarm *)sysMalloc(sizeof(Alarm));
	if( !a )
		panic("malloc failed in recordAlarm()\n");
	a->thread	= ctx;
	a->expire	= expire;
	a->data 	= data;
	a->next		= 0;
	
	pre = 0; pt = alarmList;
	if( !pt ){
		alarmList = a;
	}else{
		while( expire > pt->expire ){
			pre = pt;
			pt = pt->next;
			if( !pt ) break;
		}
		if( !pre ){
			a->next = alarmList;
			alarmList = a;
		}else{	
			a->next = pt;
			pre->next = a;
		}
	}
	alarm(alarmList->expire - cur_time);
}

static void
checkAlarm(void)
{
	Object *obj;
	thContext *ctx;
	Sync *s;
	Alarm *pt, *last;
	ulong cur_time;

	if( !alarmList ) /* liste vide */
		return;

	cur_time = getMiliSec();

	if( cur_time < alarmList->expire ){ /* il y a rien a reveiller */
		alarm(alarmList->expire - cur_time);
		return;
	}
		
	last = alarmList;

	while( last->expire <= cur_time ){
		ctx = last->thread;

		if( ctx->state == TH_WAIT_EVENT ){ /* le thread doit refaire un lock */
			obj = (Object *)last->data;
			s = &obj->s;
			rmQueue(&s->event, ctx);
			ctx->state = TH_WAIT_EVENT;
			addQueue(&s->wait, ctx);
		}else{
			addQueue(&runQ[ctx->prio], ctx);
			ctx->state = TH_READY;
		}
		pt = last->next;
		free(last);
		last = pt;
		if( !last ) break;
	}
	alarmList = last;
	if( alarmList ){
		alarm(alarmList->expire - cur_time);
	}
}

void
alarmHandler(void *u, char *str)
{
	USED(u);

	if( !strcmp(str, "alarm") ){
		ASK_SCHEDULE();
		noted(NCONT);
	}
	noted(NDFLT);
}

int 
getLock(Object *obj)
{
	Thread *t;
	thContext *ctx;
	Sync *s = &obj->s;


	if( !s->ref || s->thread_owner == current_thread ){ 
		s->ref++;
		s->thread_owner = current_thread;
		return 1;
	}

	/* object already locked */
	t = findThread(current_thread);
	ctx = (thContext *)t->private;

	rmQueue(&runQ[ctx->prio], ctx);
	ctx->state = TH_WAIT;
	addQueue(&s->wait, ctx);
	
	ASK_SCHEDULE();
	return -1;
}

void releaseLock(Object *obj)
{
	thContext *ctx, *pctx;
	Sync *s = &obj->s;

	if( s->ref <= 0 )
		panic("releaseLock(%d) for thread %d (%d)\n", obj, current_thread, s->ref);

	if( --s->ref == 0 ){
		ctx = s->wait;
		while( ctx ){
			switch(ctx->state){
				case TH_WAIT_EVENT:
					ctx->state = TH_READY_LOCK;
					ctx->data  = (int)obj;
					break;
				case TH_WAIT:
					ctx->state = TH_READY;
					break;
				default:
					panic("bad thread state in releaseLock() %d\n", ctx->state);
			}
			pctx = ctx->next;
			addQueue(&runQ[ctx->prio], ctx);
			ctx = pctx;
		}
		s->thread_owner = 0;
		s->wait = 0;
		ASK_SCHEDULE();
	}
}

/*
 * Users functions from Java API
 */

Object *
getCurrentThread(void)
{
	return current_thread;
}

void 
yieldThread(void)
{
	Thread *t = findThread(current_thread);
	thContext *ctx = (thContext *)t->private;

	rmQueue(&runQ[ctx->prio], ctx);
	addLastQueue(&runQ[ctx->prio], ctx);
	ctx->state = TH_READY;
	ASK_SCHEDULE();
}

void 
sleepThread(vlong ms)
{	
	Thread *t = findThread(current_thread);
	thContext *ctx = (thContext *)t->private;

	if( !ms )
		return;

	rmQueue(&runQ[ctx->prio], ctx);
	ctx->state = TH_SLEEP;

	recordAlarm(current_thread, (ulong)ms, 0);
	ASK_SCHEDULE();
}

int 
isAliveThread(Object *thread)
{
	Thread *t = findThread(thread);
	thContext *ctx = (thContext *)t->private;
	
	if( !ctx )
		return 0;

	if( ctx->state == TH_SUSPEND )
		return 0;
	
	return 1;
}

void 
stopThread(Object *thread, Object *th)
{
	USED(th);
	
	killThread(thread);
	ASK_SCHEDULE();
}

void
suspendThread(Object *thread)
{
	Thread *t = findThread(thread);
	thContext *ctx = (thContext *)t->private;

	rmQueue(&runQ[ctx->prio], ctx);
	ctx->state = TH_SUSPEND;
	addQueue(&suspendQ, ctx);
	ASK_SCHEDULE();
}

void
resumeThread(Object *thread)
{
	Thread *t = findThread(thread);
	thContext *ctx = (thContext *)t->private;

	rmQueue(&suspendQ, ctx);
	addQueue(&runQ[ctx->prio], ctx);
	ctx->state = TH_READY;
	ASK_SCHEDULE();
}

void
setPriorityThread(Object *thread, int prio)
{
	Thread *t = findThread(thread);
	thContext *ctx = (thContext *)t->private;


	if( !ctx ) /* le thread n'a pas commencer (start) */
		return;

	switch(ctx->state){
		case TH_RUNNING:
		case TH_READY:
			rmQueue(&runQ[ctx->prio], ctx);
			ctx->prio = prio;
			addQueue(&runQ[ctx->prio], ctx);
			ASK_SCHEDULE();
			break;
		case TH_SLEEP:
		case TH_SUSPEND:
		default:
			panic("bad thread state in setPriorityThread()\n");
	}
}

void
waitThread(Object *object, vlong timeout)
{
	Thread *t;
	thContext *ctx;
	Sync *s = &object->s;


	if( s->thread_owner != current_thread ){
		throwException(current_frame, "java/lang/IllegalMonitorStateException");
		return;
	}

	t = findThread(current_thread);
	ctx = (thContext *)t->private;

	/* enleve de la queue de priorite */
	rmQueue(&runQ[ctx->prio], ctx);
	ctx->state = TH_WAIT_EVENT;

	/* ajoute sur la queue des evts de l'objet */
	addQueue(&s->event, ctx);

	releaseLock(object);

	if( timeout )
		recordAlarm(current_thread, timeout, (int)object);

	ASK_SCHEDULE();
}


void
notifyThread(Object *object)
{
	thContext *ctx;
	Sync *s = &object->s;
	
	if( !s->event )
		return;

	ctx = s->event;
	rmQueue(&s->event, ctx);
	ctx->state = TH_WAIT_EVENT; /* indique le passage de EVENT à WAIT queue */
	addQueue(&s->wait, ctx);
	
	rmFromAlarmList(ctx);
}

void
notifyAllThread(Object *object)
{
	thContext *ctx;
	Sync *s = &object->s;
	
	if( !s->event )
		return;

	ctx = s->event;
	while( ctx ){
		rmQueue(&s->event, ctx);
		ctx->state = TH_WAIT_EVENT; /* indique le passage de EVENT à WAIT queue */
		addQueue(&s->wait, ctx);
		rmFromAlarmList(ctx);
		ctx = s->event;
	}
}

void 
startThread(Object *thread)
{
	//MethodBlock *mb;
	Frame *main_frame;
	Thread *t = findThread(thread);
	thContext *ctx;
	int n;
	MethodBlock *p = 0;

	ctx = (thContext *)sysCalloc(sizeof(thContext));
	if( !ctx )
		panic("calloc failed in startThread()\n");

	ctx->state 		= TH_READY;
	ctx->prio 		= t->priority;
	ctx->stack 		= initGlobalStack();
	ctx->thread		= thread;

	t->private 		= (int)ctx;
	addQueue(&runQ[ctx->prio], ctx);

	//print("class = %s", thread->class->r_name);

	if( strcmp(thread->class->r_name, "java/lang/Thread") ){

		for (n=thread->class->r_n_methods-1; n >= 0 ; n--){
			p = thread->class->r_methods + n;
			if( !strcmp("run()V", p->name) )
				goto found;
		}
		panic("method run not found in startThread\n");
found:
		//mb = (MethodBlock *)cur_class->r_methods + n;
		main_frame = createFrame(ctx->stack, p, 0);
	}else{
		main_frame = createFrameFromName(ctx->stack, thread->class, "run", "()V");
	}
	if( !main_frame )
		panic("method run not found in startThread\n");	
	main_frame->locals[0] = (int)thread;

	ASK_SCHEDULE();
}

/*
 * Functions to mainpulates Queues
 */

static void
addQueue(thContext **head, thContext *t)
{
	t->next = *head;
	*head = t;
}

static void
rmQueue(thContext **head, thContext *t)
{
	thContext *pt = *head;

	if( pt == t ){
		*head = t->next;
		t->next = 0;
	}else{
		while( pt->next != t )
			pt = pt->next;

		pt->next = t->next;
		t->next = 0;
	}
}

static void
addLastQueue(thContext **head, thContext *t)
{
	thContext *pt = *head;

	if( pt == 0 ){
		*head = t;
		t->next = 0;
	}else{
		while( pt->next != 0 )
			pt = pt->next;

		pt->next = t;
		t->next = 0;
	}
}


/*
 * local helper functions
 */

static Thread *
findThread(Object *thread)
{
	ClassFile *cur_class, *thread_class;
	char o_sign[256];
	int n, count;

	cur_class = thread->class;
	thread_class = getClass("java/lang/Thread");

	count = 0;
	while( cur_class != thread_class ){
		for (n=0; n < cur_class->fields_count ; n++){
			if( cur_class->fields[n].access_flag&ACC_STATIC )
				continue;
			GetCpString(o_sign, cur_class->fields[n].signature_index, cur_class);
			count += signatureSize(o_sign);
		}		
		cur_class = cur_class->r_super;
	}
	
	return (Thread *)(thread->vars + count);
}
