/* Utils functions */

#include <u.h>
#include <libc.h>
#include <lock.h>

#include "java_const.h"
#include "runtime.h"
#include "fns.h"

void
convU2( u2 *x )
{
	uchar c;
	union { u2 num; uchar n[2]; } u;
  
	u.num = *x;
	c = u.n[0];
	u.n[0] = u.n[1];
	u.n[1] = c;
	*x = u.num;
}
void 
convU4( u4 *x )
{
	u2 s1, s2;

	s1 = *x >> 16;
	s2 = *x & 0xFFFF;
	convU2(&s1); convU2(&s2);
	*x = (s2 << 16) | s1;
}

long long 
atoll( char *s ) 
{
	long long l = 0, sign = 1;

	while( *s == ' '  || *s == '\t' )
		s += 1;	/* eat spaces */

	if( *s == '-' )
		sign = -1, s++;

	while( *s >= '0' && *s <= '9' )
		l = l*10 + (*s++ - '0');
		
	return l * sign;
}

void *
sysMalloc( int size )
{
	return malloc(size);
}
void *
sysCalloc( int size )
{
	return calloc(size,1);
}
void
sysExit(int status)
{
	char buf[32];
	extern void dumpProfile(void);

	sprint(buf, "Error %d", status);
#ifdef PROFILE
	dumpProfile();
#endif
	exits(buf);
}

/*
 * Décalage à droite d'1 entier sur 64 bits
 * sans propagation du signe.
 */

void
lushr(vlong *l, int d)
{
	vlong v = 0x8000000000000000LL;

	*l >>= d;
	v >>= d-1;
	*l &= ~v;
}

long
getMiliSec(void)
{
	static int fd = -1;
	char buf[32];

	memset(buf, 0, sizeof(buf));
	if(fd < 0)
		fd = open("/dev/msec", OREAD|OCEXEC);
	if(fd >= 0) {
		seek(fd, 0, 0);
		read(fd, buf, sizeof(buf));
	}

	return atol(buf);
}

int
getNbPath(char * path)
{
	int n = 0;
	char *p = path;
	
	while( (p=strstr(p, " ")) )
		n++, p++;
	
	return n+1; /* attention ne doit pas se terminer par ' ' */
}

void
getPath(char *buf, char *path, int nb)
{
	int n = 0;
	char *p, *pre;

	p = path;
	while( n++ < nb ){
		while( *p != ' ' && *p )
			p++;
		p += 1;
	}
	pre = p;
	while( *p != ' ' && *p )
			p++;

	strncpy(buf, pre, p-pre);
	buf[p-pre] = 0;
}

int 
truncName(char *name)
{
	char *pt;

	pt = name + strlen(name) - 1;
	
	while( *pt != '/' && pt > name) 
		pt--;
	if( *pt == '/' )
		pt++;
	
	if( strlen(pt) > NAMELEN-1 ){
		strcpy(pt + 23, "..a");
		return 1;
	}
	return 0;
}

void
slash2point(char *buf)
{
	while( *buf ){	
		if( *buf == '/' )
			*buf = '.';
		buf++;
	}
}

void
point2slash(char *buf)
{
	while( *buf ){	
		if( *buf == '.' )
			*buf = '/';
		buf++;
	}
}

//#ifdef DBG_1
void
printAccessFlag( int flag )
{
	chat("< ");
	if( flag&ACC_PUBLIC )
		chat("PUBLIC ");
	if( flag&ACC_PRIVATE )
		chat("PRIVATE ");
	if( flag&ACC_PROTECTED )
		chat("PROTECTED ");
	if( flag&ACC_STATIC )
		chat("STATIC ");
	if( flag&ACC_FINAL )
		chat("FINAL ");
	if( flag&ACC_SYNCHRONIZED )
		chat("SYNCHRONIZED ");
	if( flag&ACC_VOLATILE )
		chat("VOLATILE ");
	if( flag&ACC_TRANSIENT )
		chat("TRANSIENT ");
	if( flag&ACC_NATIVE )
		chat("NATIVE ");
	if( flag&ACC_INTERFACE )
		chat("INTERFACE ");
	if( flag&ACC_ABSTRACT )
		chat("ABSTRACT ");
	chat(">");
}

void
dumpFrame( Frame *f )
{
	int i;
	
	chat("*** Frame\n");
	chat("name = %s.%s\n", f->class->r_name, f->name);
	for(i=0 ; i < FRAME_MAX_LOCALS(f) ; i++)
		chat("locals[%d] = %d, ", i, f->locals[i]);
	chat("\n");
	for(i=0 ; i < FRAME_MAX_STACK(f) ; i++)
		chat("stack[%d] = %d, ", i, f->opstack[i]);
	chat("\n");
	chat("stack pointer = %d\n", f->opstack_pt);
}

void 
dumpMethodBlock( MethodBlock *m )
{
	chat("*** Method block\n");
	chat("class name = %s\n", m->class->r_name);
	chat("nargs = %d\n", m->nargs);
	chat("max_locals = %d :: max_stack = %d\n", 
							MB_MAX_LOCALS(m), MB_MAX_STACK(m));
}

void 
dumpObject( Object *ob )
{
	int i, count;

	if( IS_OBJECT(ob) ){
		chat("*** Object\n");
		chat("class = %s\n", ob->class->r_name);
	}else if( IS_ARRAY(ob) ){
		chat("*** Primitive Array\n");	
		switch(ob->type){
			case T_CLASS:
				chat("type = T_CLASS impossible !!!!\n");
				break;
			case T_TAB:
				chat("type = T_TAB impossible !!!!\n");
				break;
			case T_STRING:
				chat("type = T_STRING impossible !!!!\n");
				break;
			case T_VOID:
				chat("type = T_VOID impossible !!!!\n");
				break;
			case T_BOOLEAN:
				chat("type = T_BOOLEAN\n");
				break;
			case T_CHAR:
				chat("type = T_CHAR\n");
				break;
			case T_FLOAT:
				chat("type = T_FLOAT\n");
				break;
			case T_DOUBLE:
				chat("type = T_DOUBLE\n");
				break;
			case T_BYTE:
				chat("type = T_BYTE\n");
				break;
			case T_SHORT:
				chat("type = T_SHORT\n");
				break;
			case T_INT:
				chat("type = T_INT\n");
				break;
			case T_LONG:
				chat("type = T_LONG\n");
				break;
		default:
				chat("unknown type.\n");
		}
	}else if( IS_ARRAY_OBJECT(ob) ){
		chat("*** Array Object\n");	
	}

	chat("size = %d\n", ob->vars_size);
	count = (ob->vars_size > 100 ) ? 100 : ob->vars_size;
	for(i=0 ; i < count ; i++)
		chat("vars[%d] = %d ", i, ob->vars[i]);
	chat("\n");
}




void 
dumpString( Object *ob )
{
	int i;

	chat("*** Object\n");
	if( ob->class )
		print("class name = %s\n", ob->class->r_name);
	chat("vars_size = %d\n", ob->vars_size);
	if(0)
	for(i=0 ; i < ob->vars_size ; i++)
		chat("vars[%d] = %d ", i, ob->vars[i]);
	chat("\n");
}
//#endif
